# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the base class for parameterised phase diagrams. """

import tqdm
from matador.hull import PhaseDiagram, QueryConvexHull
from matador.utils.cursor_utils import filter_cursor_by_chempots, recursive_get, recursive_set, set_cursor_from_array
from matador.utils.chem_utils import get_formation_energy, get_root_source, get_formula_from_stoich


class EnsembleHull(QueryConvexHull):
    """ Class to create and store an ensemble of composition vs energy convex
    hulls from cursor data. The variable energies must be stored under a given
    key, e.g. `doc['_beef'][energy_key][beef_index]`, as specified by init.

    Data must be stored in the following way under each document in cursor::

        {...,
         data_key: {
            parameter_key: "<list of parameter values>",
            energy_key: "<list of energies at parameter values>",
         },
        }

    Hull data will be stored as arrays per document under `doc[data_key]['hull_distance']`
    and ``doc[data_key]['formation_' + energy_key]``.

    Inherits the attributes of matador.hull.QueryConvexHull, with many set to
    None.

    Attributes:
        phase_diagrams (list of :obj:`matador.hull.PhaseDiagram`): list of phase diagram
            objects for each parameter value.

    """
    def __init__(self, cursor, data_key, energy_key='enthalpy_per_atom', chempot_energy_key=None,
                 num_samples=None, parameter_key=None, species=None, voltage=False, verbosity=None, **kwargs):
        """ Initialise EnsembleHull from a cursor, with other keywords
        following QueryConvexHull.

        Parameters:
            cursor (list[dict]): list of matador documents containing
                variable parameter data for energies.
            data_key (str): the key under which all parameter data is
                stored to the variable parameter, e.g. `_beef` or `_temperature`.

        Keyword arguments:
            energy_key (str): the key under `parameter_key` to use to create
                the hulls.
            chempot_energy_key (str): the key used to create the first convex hull.
            parameter_key (str): the key pertaining to the variable parameter
                itself, e.g. `temperature` or `thetas`.
            num_samples (int): use up to this many samples in creating the hull.
            species (list[str]): list of elements/chempots to use, in
                the desired order.
            voltage (bool): whether or not to compute voltage curves.
            plot_kwargs (dict): arguments to pass to plot_hull function.
            kwargs (dict): other arguments to pass to QueryConvexHull.

        """
        # sometimes the first hull needs to be made with a different key
        if chempot_energy_key is not None:
            self.chempot_energy_key = chempot_energy_key
        else:
            self.chempot_energy_key = energy_key

        super().__init__(cursor=cursor,
                         energy_key=self.chempot_energy_key,
                         species=species,
                         voltage=voltage,
                         no_plot=True,
                         lazy=False,
                         **kwargs)

        self.energy_key = energy_key

        if self.phase_diagram is None:
            del self.phase_diagram
        if self.hull_dist is None:
            del self.hull_dist

        self.from_cursor = True
        self.verbosity = verbosity
        # set up relative keys
        self.formation_key = 'formation_' + self.energy_key
        self.data_key = data_key
        self.parameter_key = parameter_key

        if self.parameter_key is None:
            self._parameter_keys = None
        else:
            self._parameter_keys = [self.data_key] + [parameter_key]
        self._formation_keys = [self.data_key] + [self.formation_key]
        self._hulldist_keys = [self.data_key] + ['hull_distance']
        self._energy_keys = [self.data_key] + [self.energy_key]

        self.phase_diagrams = []

        self.set_chempots(energy_key=self.chempot_energy_key)
        self.cursor = filter_cursor_by_chempots(self.species, self.cursor)
        self.cursor = sorted(self.cursor, key=lambda doc: (recursive_get(doc, self.chempot_energy_key), doc['concentration']))

        if self.parameter_key is None:
            parameter_iterable = recursive_get(self.chempot_cursor[0], self._energy_keys)
        else:
            parameter_iterable = recursive_get(self.chempot_cursor[0], self._parameter_keys)

        print('Found {} entries under data key: {}.'.format(len(parameter_iterable), self.data_key))

        # allocate formation energy and hull distance arrays
        for ind, doc in enumerate(self.cursor):
            recursive_set(doc, self._formation_keys, [None] * len(recursive_get(doc, self._energy_keys)))
            recursive_set(doc, self._hulldist_keys, [None] * len(recursive_get(doc, self._energy_keys)))

        n_hulls = len(parameter_iterable)
        if num_samples is not None:
            parameter_iterable = parameter_iterable[:num_samples]
            print('Using {} out of {} possible phase diagrams.'.format(num_samples, n_hulls))
        else:
            num_samples = n_hulls

        for param_ind, parameter in enumerate(tqdm.tqdm(parameter_iterable)):
            for ind, doc in enumerate(self.cursor):
                if self.parameter_key is not None:
                    assert recursive_get(doc, self._parameter_keys + [param_ind]) == parameter

                formation_energy = get_formation_energy(self.chempot_cursor, doc,
                                                        energy_key=self._energy_keys + [param_ind])
                recursive_set(self.cursor[ind], self._formation_keys + [param_ind], formation_energy)
            self.phase_diagrams.append(PhaseDiagram(self.cursor,
                                                    self._formation_keys + [param_ind],
                                                    self._dimension))
            set_cursor_from_array(self.cursor,
                                  self.phase_diagrams[-1].hull_dist,
                                  self._hulldist_keys + [param_ind])

        self.stability_histogram = self.generate_stability_statistics()

    def generate_stability_statistics(self, group_by='structure'):
        """ Creates a histogram that counts how many times each structure
        is found to be stable in the ensemble.

        Keyword arguments:
            group_by (str): either 'structure' or 'formula' for bar groupings.

        """
        from collections import defaultdict
        histogram = defaultdict(int)
        for pd in self.phase_diagrams:
            for doc in pd.stable_structures:
                if group_by == 'formula':
                    histogram[get_formula_from_stoich(doc['stoichiometry'])] += 1
                else:
                    histogram[get_root_source(doc)] += 1
        return histogram

    def plot_hull(self, **kwargs):
        """ Hull plot helper function. """
        from matador.plotting.hull_plotting import plot_ensemble_hull
        return plot_ensemble_hull(self, self.data_key, formation_energy_key=self.formation_key, **kwargs)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the CastepSpectralWorkflow, which performs
spectral calculations with CASTEP in multiple steps:

1. Performs a singlepoint calculation (if check file is not found).
2. If the `spectral_kpoints_mp_spacing` keyword is found, interpolate
   wavefunction to DOS grid.
   - If an OptaDOS input file (.odi) with the root seedname
     is found, run OptaDOS on the resulting density of states.
3. If `spectral_kpoints_path_spacing` keyword is present, create
   a bandstructure on the SeeK-path generated path.

"""


import os
import copy
import glob
import logging
from matador.workflows import Workflow
from matador.scrapers import arbitrary2dict

LOG = logging.getLogger('run3')


def castep_full_spectral(relaxer, calc_doc, seed, **kwargs):
    """ Perform a "full" spectral calculation on a system, i.e. first
    perform an SCF then interpolate to different kpoint paths/grids to
    form DOS and dispersions. Optionally use OptaDOS for post-processing
    of DOS.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepSpectralWorkflow(relaxer, calc_doc, seed)
    return workflow.success


class CastepSpectralWorkflow(Workflow):
    """ Perform a "full" spectral calculation on a system, i.e. first
    perform an SCF then interpolate to different kpoint paths/grids to
    form DOS and dispersions. Optionally use OptaDOS for post-processing
    of DOS.

    Attributes:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.

    """
    def preprocess(self):
        """ Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """
        # default todo
        todo = {'scf': True, 'dos': False, 'pdos': False, 'broadening': False, 'dispersion': False, 'pdis': False}
        # definition of steps and names
        steps = {'scf': castep_spectral_scf,
                 'dos': castep_spectral_dos,
                 'pdos': optados_pdos,
                 'broadening': optados_dos_broadening,
                 'dispersion': castep_spectral_dispersion,
                 'pdis': optados_pdispersion}

        exts = {
            'scf': {
                'input': ['.cell', '.param'],
                'output': ['.castep', '.*err', '-out.cell']
            },
            'dos': {
                'input': ['.cell', '.param'],
                'output': ['.castep', '.bands', '.pdos_bin', '.dome_bin', '.*err', '-out.cell']
            },
            'dispersion': {
                'input': ['.cell', '.param'],
                'output': ['.castep', '.bands', '.pdos_bin', '.dome_bin', '.*err', '-out.cell']
            },
            'pdis': {
                'input': ['.odi', '.pdos_bin'],
                'output': ['.odo', '.*err']
            },
            'pdos': {
                'input': ['.odi', '.pdos_bin', '.dome_bin'],
                'output': ['.odo', '.*err']
            },
            'broadening': {
                'input': ['.odi', '.pdos_bin', '.dome_bin'],
                'output': ['.odo', '.*err']
            }
        }

        if os.path.isfile(self.seed + '.check'):
            LOG.info('Found {}.check, so skipping initial SCF.'.format(self.seed))
            todo['scf'] = False

        if (
                (
                    'spectral_kpoints_path' in self.calc_doc or
                    'spectral_kpoints_list' in self.calc_doc or
                    'spectral_kpoints_path_spacing' in self.calc_doc or
                    self.calc_doc.get('spectral_task', '').lower() == 'bandstructure'
                )
        ):
            todo['dispersion'] = not os.path.isfile(self.seed + '.bands_dispersion')

        if ('spectral_kpoints_mp_spacing' in self.calc_doc or
                self.calc_doc.get('spectral_task', '').lower() == 'dos'):
            todo['dos'] = not os.path.isfile(self.seed + '.bands_dos')

        odi_fname = _get_optados_fname(self.seed)
        if odi_fname is not None:
            odi_dict, _ = arbitrary2dict(odi_fname)
            if todo['dispersion']:
                todo['pdis'] = 'pdispersion' in odi_dict
            if todo['dos']:
                todo['broadening'] = 'broadening' in odi_dict
                todo['pdos'] = 'pdos' in odi_dict

        for key in todo:
            if todo[key]:
                self.add_step(steps[key], key,
                              input_exts=exts[key].get('input'),
                              output_exts=exts[key].get('output'))

        if self.relaxer.run3_settings.get('run3_settings') is not None:
            settings = self.relaxer.kwargs.get('run3_settings')
            # check that relaxer.exec was not overriden at cmd-line, then check settings file
            if settings.get('castep_executable') is not None and self.relaxer.executable == 'castep':
                self.castep_executable = settings.get('castep_executable', 'castep')
                self.relaxer.executable = self.castep_executable
            if settings.get('optados_executable') is not None:
                self.optados_executable = settings.get('optados_executable', 'optados')
                self.relaxer.optados_executable = self.optados_executable

        # if not using a user-requested path, use seekpath and spglib
        # to reduce to primitive and use consistent path
        if 'spectral_kpoints_list' not in self.calc_doc and 'spectral_kpoints_path' not in self.calc_doc:
            from matador.utils.cell_utils import cart2abc
            prim_doc, kpt_path = self.relaxer.get_seekpath_compliant_input(
                self.calc_doc, self.calc_doc.get('spectral_kpoints_path_spacing', 0.05))
            self.calc_doc.update(prim_doc)
            self.calc_doc['lattice_abc'] = cart2abc(self.calc_doc['lattice_cart'])
            if todo['dispersion']:
                self.calc_doc['spectral_kpoints_list'] = kpt_path
        elif todo['dispersion'] and 'spectral_kpoints_path' in self.calc_doc:
            self._user_defined_kpt_path = True
            LOG.warning('Using user-defined k-point path for all structures.')
            self.calc_doc['spectral_kpoints_path_spacing'] = self.calc_doc.get('spectral_kpoints_path_spacing', 0.05)

        if todo['dos']:
            self.calc_doc['spectral_kpoints_mp_spacing'] = self.calc_doc.get('spectral_kpoints_mp_spacing', 0.05)

        # always use continuation
        self.calc_doc['continuation'] = 'default'

        LOG.info('Preprocessing completed: run3 spectral options {}'.format(todo))


def castep_spectral_scf(relaxer, calc_doc, seed):
    """ Run a singleshot SCF calculation.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP spectral SCF...')
    scf_doc = copy.deepcopy(calc_doc)
    scf_doc['write_checkpoint'] = 'ALL'
    scf_doc['task'] = 'singlepoint'
    scf_doc['write_cell_structure'] = True
    if 'spectral_task' in scf_doc:
        del scf_doc['spectral_task']

    required = []
    forbidden = ['spectral_task',
                 'spectral_kpoints_list',
                 'spectral_kpoints_path',
                 'spectral_kpoints_mp_spacing',
                 'spectral_kpoints_path_spacing']

    relaxer.validate_calc_doc(scf_doc, required, forbidden)

    return relaxer.run_castep_singleshot(scf_doc, seed, keep=True, intermediate=True)


def castep_spectral_dos(relaxer, calc_doc, seed):
    """ Runs a DOS interpolation on top of a completed SCF. If a single
    .odi file is found, run OptaDOS on the resulting DOS.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP spectral DOS calculation...')

    dos_doc = copy.deepcopy(calc_doc)
    dos_doc['task'] = 'spectral'
    dos_doc['spectral_task'] = 'dos'
    # disable checkpointing for BS/DOS by default, leaving just SCF
    dos_doc['write_checkpoint'] = 'none'
    dos_doc['write_cell_structure'] = True
    dos_doc['pdos_calculate_weights'] = True

    required = ['spectral_kpoints_mp_spacing']
    forbidden = ['spectral_kpoints_list',
                 'spectral_kpoints_path',
                 'spectral_kpoints_path_spacing']

    relaxer.validate_calc_doc(dos_doc, required, forbidden)
    success = relaxer.run_castep_singleshot(dos_doc, seed, keep=True, intermediate=True)

    return success


def castep_spectral_dispersion(relaxer, calc_doc, seed):
    """ Runs a dispersion interpolation on top of a completed SCF calculation,
    optionally running orbitals2bands and OptaDOS projected dispersion.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info('Performing CASTEP spectral dispersion calculation...')
    disp_doc = copy.deepcopy(calc_doc)
    disp_doc['task'] = 'spectral'
    disp_doc['spectral_task'] = 'bandstructure'

    # disable checkpointing for BS/DOS by default, leaving just SCF
    disp_doc['write_checkpoint'] = 'none'
    disp_doc['pdos_calculate_weights'] = True
    disp_doc['write_cell_structure'] = True
    disp_doc['continuation'] = 'default'

    required = []
    forbidden = ['spectral_kpoints_mp_spacing']

    relaxer.validate_calc_doc(disp_doc, required, forbidden)
    success = relaxer.run_castep_singleshot(disp_doc, seed, keep=True, intermediate=True)

    if disp_doc.get('write_orbitals'):
        LOG.info('Planning to call orbitals2bands...')

        _cache_executable = copy.deepcopy(relaxer.executable)
        _cache_core = copy.deepcopy(relaxer.ncores)
        relaxer.ncores = 1
        relaxer.executable = 'orbitals2bands'
        try:
            success = relaxer.run_generic(intermediate=True, mv_bad_on_failure=False)
        except Exception as exc:
            relaxer.executable = _cache_executable
            relaxer.ncores = _cache_core
            LOG.warning('Failed to call orbitals2bands, with error: {}'.format(exc))

        relaxer.ncores = _cache_core
        relaxer.executable = _cache_executable

    return success


def optados_pdos(relaxer, _, seed):
    """ Run an OptaDOS projected-DOS.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """

    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)

        odi_dict['task'] = 'pdos'
        if 'pdispersion' in odi_dict:
            del odi_dict['pdispersion']

        LOG.info('Performing OptaDOS pDOS calculation with parameters from {}'.format(odi_fname))
        success = _run_optados(relaxer, odi_dict, seed, suffix='dos')
        return success

    return None


def optados_dos_broadening(relaxer, _, seed):
    """ Run an OptaDOS total DOS broadening.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """

    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)
        # if broadening keyword is present, try to run a normal DOS
        odi_dict['task'] = 'dos'
        if 'pdos' in odi_dict:
            del odi_dict['pdos']
        if 'pdispersion' in odi_dict:
            del odi_dict['pdispersion']

        LOG.info('Performing OptaDOS DOS broadening with parameters from {}'.format(odi_fname))
        return _run_optados(relaxer, odi_dict, seed, suffix='dos')

    return None


def optados_pdispersion(relaxer, _, seed):
    """ Runs an OptaDOS projected dispersion calculation.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """
    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)
        # if pdispersion keyword is present, try to run a pdis
        odi_dict['task'] = 'pdispersion'
        if 'pdos' in odi_dict:
            del odi_dict['pdos']

        LOG.info('Performing OptaDOS pDIS calculation with parameters from {}'.format(odi_fname))
        return _run_optados(relaxer, odi_dict, seed, suffix='dispersion')

    return None


def _run_optados(relaxer, odi_dict, seed, suffix=None):
    """ Run OptaDOS with given relaxer object, parameters and seed, adjusting
    the number of cores and the executable to call, then restoring them after.

    Parameters:
        relaxer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        odi_dict (dict): the OptaDOS parameters to write to file.
        seed (str): root filename of structure.

    Keyword arguments:
        suffix (str): either 'dos' or 'dispersion' for backup/restore.
        executable (str): optados executable to call.

    """

    from matador.export import doc2arbitrary
    odi_path = '{}.odi'.format(seed)
    if relaxer.compute_dir is not None:
        odi_path = relaxer.compute_dir + '/' + odi_path
    doc2arbitrary(odi_dict, odi_path, overwrite=True)

    if suffix is not None:
        _get_correct_files_for_optados(seed, suffix=suffix)

    _cache_executable = copy.deepcopy(relaxer.executable)
    _cache_core = copy.deepcopy(relaxer.ncores)
    _cache_nodes = copy.deepcopy(relaxer.nnodes)
    relaxer.ncores = 1
    relaxer.nnodes = 1
    relaxer.executable = relaxer.optados_executable
    success = False

    try:
        success = relaxer.run_generic(intermediate=True, mv_bad_on_failure=False)
    except Exception as exc:
        LOG.warning('Failed to call optados with error: {}'.format(exc))

    relaxer.ncores = _cache_core
    relaxer.nnodes = _cache_nodes
    relaxer.executable = _cache_executable
    if suffix is not None:
        _get_correct_files_for_optados(seed, suffix='bak')

    return success


def _get_optados_fname(seed):
    """ Get the most likely OptaDOS input file name, which here means either
    only existing one, or the shortest one.

    Parameters:
        seed (str): seedname for the calculation.

    Returns:
        str: the OptaDOS input filename.

    """
    odi_fname = None
    if os.path.isfile(seed + '.odi'):
        os.remove(seed + '.odi')

    if glob.glob('*.odi'):
        # dodginess: choose the odi file with the shortest name...
        shortest_fname = None
        for fname in glob.glob('*.odi'):
            if shortest_fname is None or len(fname) < len(shortest_fname):
                shortest_fname = fname
        odi_fname = shortest_fname

    return odi_fname


def _get_correct_files_for_optados(seed, suffix=None):
    """ If e.g. dispersion and DOS calculations were run previously, but
    it is unclear which exists as the current <seed>.<ext>, try to copy
    the old <seed>.<ext>_dispersion/dos files to the correct place.

    Parameters:
        seed (str): seedname for the calculation.

    Keyword arguments:
        suffix (str): either 'dos' or 'dispersion', or 'bak' to get old files back.

    """
    import shutil
    LOG.debug('Getting files for OptaDOS: {} {}'.format(seed, suffix))
    if suffix is None:
        return
    files_to_cache = ['.pdos_bin', '.dome_bin', '-out.cell', '.bands', '.cell', '.param']
    for ext in files_to_cache:
        old_file = '{}{}_{}'.format(seed, ext, suffix)
        current_file = '{}{}'.format(seed, ext)
        if os.path.isfile(current_file):
            if suffix != 'bak':
                backup_file = '{}{}_{}'.format(seed, ext, 'bak')
                shutil.copy2(current_file, backup_file)
            os.remove(current_file)
        if os.path.isfile(old_file):
            LOG.debug('Copying {} to {}'.format(old_file, current_file))
            shutil.copy2(old_file, current_file)
            if suffix == 'bak':
                os.remove(old_file)



""" Test battery routines. """

import unittest

import json
from os.path import realpath
from glob import glob

import numpy as np

from matador.battery import Electrode
from matador.hull import QueryConvexHull
from matador.scrapers.castep_scrapers import res2dict
from matador.utils.chem_utils import get_concentration

# grab abs path for accessing test data
REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"


class VoltageTest(unittest.TestCase):
    """ Test Electrode voltage functionality. """

    def test_average_voltage(self):
        capacities = [0, 10, 100]
        voltages = [1, 1, 1]
        self.assertEqual(Electrode.calculate_average_voltage(capacities, voltages), 1.0)

        capacities = [0, 10, 100, np.nan]
        voltages = [1, 1, 1, 0]
        self.assertEqual(Electrode.calculate_average_voltage(capacities, voltages), 1.0)

        capacities = [0, 100, 1000]
        voltages = [1, 1, 0.1]
        self.assertEqual(
            Electrode.calculate_average_voltage(capacities, voltages), 0.19
        )

        capacities = [0, 100, 1000, np.nan]
        voltages = [1, 1, 0.1, 0]
        self.assertEqual(
            Electrode.calculate_average_voltage(capacities, voltages), 0.19
        )

    def test_binary_voltage(self):
        """ Test simple binary voltage curve. """
        hull_cursor = []
        test_Q = np.loadtxt(REAL_PATH + "data/voltage_data/LiAs_Q.dat")
        test_V = np.loadtxt(REAL_PATH + "data/voltage_data/LiAs_V.dat")
        for i in range(5):
            with open(
                REAL_PATH + "data/voltage_data/hull_data" + str(i) + ".json"
            ) as f:
                hull_cursor.append(json.load(f))
        hull = QueryConvexHull(
            cursor=hull_cursor, species=["Li", "As"], voltage=True, no_plot=True
        )
        self.assertTrue(len(hull.voltage_data) == 1)
        np.testing.assert_array_almost_equal(
            hull.voltage_data[0].voltages, test_V, decimal=5, verbose=True
        )
        np.testing.assert_array_almost_equal(
            hull.voltage_data[0].capacities, test_Q, decimal=5
        )
        self.assertAlmostEqual(
            hull.voltage_data[0].average_voltage, 0.949184, places=3
        )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_multivalent_binary_voltage(self):
        """ Test simple binary voltage curve. """
        hull_cursor = []
        test_Q = np.loadtxt(REAL_PATH + "data/voltage_data/LiAs_Q.dat")
        test_V = np.loadtxt(REAL_PATH + "data/voltage_data/LiAs_V.dat")
        for i in range(5):
            with open(
                REAL_PATH + "data/voltage_data/hull_data" + str(i) + ".json"
            ) as f:
                flines = f.readlines()
                # replace all Li with Mg
                flines = [line.replace("Li", "Mg") for line in flines]
                hull_cursor.append(json.loads("\n".join(flines)))
        hull = QueryConvexHull(
            cursor=hull_cursor, species=["Mg", "As"], voltage=True, no_plot=True
        )

        self.assertEqual(len(hull.voltage_data), 1)
        np.testing.assert_array_almost_equal(
            hull.voltage_data[0].voltages, 2 * test_V, decimal=5, verbose=True
        )
        np.testing.assert_array_almost_equal(
            hull.voltage_data[0].capacities, test_Q, decimal=5
        )
        self.assertAlmostEqual(
            hull.voltage_data[0].average_voltage, 2 * 0.949184, places=3
        )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_binary_voltage_mayo(self):
        """ Test binary voltages from cursor for Mayo et al,
        DOI: 10.1021/acs.chemmater.5b04208.

        """
        res_list = glob(REAL_PATH + "data/hull-LiP-mdm_chem_mater/*.res")
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(
            cursor=cursor, elements=["Li", "P"], no_plot=True, voltage=True
        )
        for profile in hull.voltage_data:
            self.assertEqual(
                len(profile.voltages), len(profile.capacities)
            )
        LiP_voltage_curve = np.loadtxt(
            REAL_PATH + "data/LiP_voltage.csv", delimiter=","
        )
        self.assertTrue(len(hull.voltage_data) == 1)
        np.testing.assert_allclose(
            hull.voltage_data[0].voltages,
            LiP_voltage_curve[:, 1],
            verbose=True,
            rtol=1e-4,
        )
        np.testing.assert_allclose(
            hull.voltage_data[0].capacities, LiP_voltage_curve[:, 0], verbose=True, rtol=1e-4
        )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage(self):
        """ Test ternary voltages from cursor. """
        # test data from LiSnS
        res_list = glob(REAL_PATH + "data/hull-LiSnS/*.res")
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["Li", "Sn", "S"],
            no_plot=True,
            pathways=True,
            voltage=True,
            debug=True,
        )
        pin = np.array(
            [
                [2, 0, 0, -380.071],
                [0, 2, 4, -1305.0911],
                [2, 0, 1, -661.985],
                [6, 2, 0, -1333.940],
                [16, 4, 16, -7906.417],
                [4, 4, 0, -1144.827],
                [0, 4, 4, -1497.881],
                [0, 1, 0, -95.532],
                [0, 0, 48, -13343.805],
            ]
        )
        tot = pin[:, 0] + pin[:, 1] + pin[:, 2]
        points = pin / tot[:, None]

        voltage_data = [
            np.asarray(
                [
                    1.9415250000000697,
                    1.9415250000000697,
                    1.8750000000001705,
                    1.4878749999999741,
                    0.63925000000000409,
                    0.34612500000000068,
                    0.0,
                ]
            ),
            np.asarray(
                [
                    1.4878749999999741,
                    1.4878749999999741,
                    0.63925000000000409,
                    0.34612500000000068,
                    0.0,
                ]
            ),
        ]

        Q_data = [
            np.array([0, 195, 293, 586, 733, 1026, np.NaN]),
            np.array([0, 356, 533, 889, np.NaN]),
        ]

        points = np.delete(points, 2, axis=1)

        self.assertEqual(len(hull.voltage_data), len(Q_data))
        for i in range(len(hull.voltage_data)):
            np.testing.assert_array_almost_equal(
                hull.voltage_data[i].voltages, voltage_data[i], decimal=3
            )
            np.testing.assert_array_almost_equal(
                hull.voltage_data[i].capacities, Q_data[i], decimal=0
            )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage_with_one_two_phase_region(self):
        """ Test ternary voltages with awkward two-phase region. """
        # load old hull then rejig it to go through a ternary phase
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        cursor = [
            doc
            for doc in cursor
            if (
                doc["stoichiometry"] != [["P", 3], ["Sn", 4]]
                and doc["stoichiometry"] != [["P", 3], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 3], ["P", 7]]
                and doc["stoichiometry"] != [["K", 1], ["P", 7]]
                and doc["stoichiometry"] != [["K", 2], ["P", 3]]
                and doc["stoichiometry"] != [["K", 8], ["P", 4], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 1], ["P", 2], ["Sn", 2]]
                and doc["stoichiometry"] != [["K", 1], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 4], ["Sn", 9]]
                and doc["stoichiometry"] != [["K", 5], ["P", 4]]
                and doc["stoichiometry"] != [["P", 2], ["Sn", 1]]
            )
        ]
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=True,
            pathways=True,
            voltage=True,
        )
        np.testing.assert_array_almost_equal(
            np.asarray(hull.voltage_data[0].voltages),
            np.asarray([1.0229, 1.0229, 0.2676, 0.000]),
            decimal=3,
        )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage_with_two_two_phase_regions(self):
        """ Test ternary voltages with two awkward two-phase regions. """
        # load old hull then rejig it to go through a ternary phase
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        cursor = [
            doc
            for doc in cursor
            if (
                doc["stoichiometry"] != [["P", 3], ["Sn", 4]]
                and doc["stoichiometry"] != [["P", 3], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 3], ["P", 7]]
                and doc["stoichiometry"] != [["K", 1], ["P", 7]]
                and doc["stoichiometry"] != [["K", 2], ["P", 3]]
                and doc["stoichiometry"] != [["K", 8], ["P", 4], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 1], ["Sn", 1]]
                and doc["stoichiometry"] != [["K", 4], ["Sn", 9]]
                and doc["stoichiometry"] != [["K", 5], ["P", 4]]
                and doc["stoichiometry"] != [["P", 2], ["Sn", 1]]
            )
        ]
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=True,
            pathways=True,
            voltage=True,
        )
        np.testing.assert_array_almost_equal(
            np.asarray(hull.voltage_data[0].voltages),
            np.asarray([1.1845, 1.1845, 0.8612, 0.2676, 0.000]),
            decimal=3,
        )
        self.assertAlmostEqual(hull.voltage_data[0].capacities[-2], 425.7847612, places=5)
        self.assertAlmostEqual(
            hull.voltage_data[0].average_voltage, 0.58523, places=4
        )
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage_with_exclusively_two_phase_regions(self):
        """ Test ternary voltages exclusively awkward two-phase regions. """
        # load old hull then rejig it to go through a ternary phase
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        cursor = [
            doc
            for doc in cursor
            if (
                doc["stoichiometry"] == [["P", 1], ["Sn", 1]]
                or doc["stoichiometry"] == [["K", 1], ["P", 1], ["Sn", 1]]
                or doc["stoichiometry"] == [["P", 1]]
                or doc["stoichiometry"] == [["K", 1]]
                or doc["stoichiometry"] == [["Sn", 1]]
            )
        ]

        hull = QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=True,
            pathways=True,
            voltage=True,
        )
        self.assertEqual(len(hull.voltage_data), 1)
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage_with_single_phase_region(self):
        """ Test ternary voltages with single-phase regions. """
        # load old hull then rejig it to go through a ternary phase
        cursor = res2dict(REAL_PATH + "data/hull-LiSiP/*.res")[0]
        hull = QueryConvexHull(
            cursor=cursor, species="LiSiP", no_plot=True, voltage=True
        )
        np.testing.assert_array_almost_equal(
            np.asarray(hull.voltage_data[0].voltages),
            np.asarray(
                [
                    1.1683,
                    1.1683,
                    1.0759,
                    0.7983,
                    0.6447,
                    0.3726,
                    0.3394,
                    0.1995,
                    0.1570,
                    0.1113,
                    0.1041,
                    0.0000,
                ]
            ),
            decimal=3,
        )
        self.assertEqual(len(hull.voltage_data[0].capacities), 12)
        self.assertEqual(len(hull.voltage_data[1].capacities), 11)
        self._check_voltages_match_capacities(hull.voltage_data)

    def test_ternary_voltage_problematic(self):
        """ Test for NaSnP voltages which triggered a bug in the capacity
        calculation.

        """
        cursor = res2dict(REAL_PATH + "data/voltage_data/voltage-NaSnP/*.res")[0]
        hull = QueryConvexHull(
            cursor=cursor, species="NaSnP", no_plot=True, voltage=True, volume=True
        )
        self._check_voltages_match_capacities(hull.voltage_data)
        self.assertEqual(len(hull.voltage_data), 2)
        self.assertEqual(len(hull.voltage_data[0].capacities), 13)
        self.assertEqual(len(hull.voltage_data[1].capacities), 12)

    def test_ternary_voltage_problematic_with_crystal_models(self):
        """ Test for NaSnP voltages which triggered a bug in the capacity
        calculation.

        """
        cursor = res2dict(
            REAL_PATH + "data/voltage_data/voltage-NaSnP/*.res", as_model=True
        )[0]
        hull = QueryConvexHull(
            cursor=cursor, species="NaSnP", no_plot=True, voltage=True, volume=True
        )
        self._check_voltages_match_capacities(hull.voltage_data)
        self.assertEqual(len(hull.voltage_data), 2)
        self.assertEqual(len(hull.voltage_data[0].capacities), 13)
        self.assertEqual(len(hull.voltage_data[1].capacities), 12)

    def test_angelas_awkward_voltage(self):
        """ Test a particular example of Angela's awkward ternary voltages. """
        # test data from NaFeP
        res_list = glob(REAL_PATH + "data/hull-NaFeP-afh41_new_Na+Fe+P/*.res")
        self.assertEqual(
            len(res_list),
            16,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(
            cursor=cursor, elements=["Na", "Fe", "P"], no_plot=True, voltage=True
        )
        self.assertEqual(len(hull.voltage_data[0].voltages), 8)
        self.assertEqual(len(hull.voltage_data[1].voltages), 5)
        self.assertEqual(len(hull.voltage_data[2].voltages), 3)

        self._check_voltages_match_capacities(hull.voltage_data)

    def _check_voltages_match_capacities(self, voltage_data):
        for profile in voltage_data:
            self.assertEqual(
                len(profile.voltages), len(profile.capacities)
            )
            self.assertTrue(np.isnan(profile.capacities[-1]))
            self.assertEqual(profile.voltages[-1], 0.0)


class VolumeTest(unittest.TestCase):
    """ Test simple binary volume curve. """

    def test_binary_volume_curve(self):
        """ Test simple binary volume curve. """
        res_list = glob(REAL_PATH + "data/hull-LiP-mdm_chem_mater/*.res")
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(cursor=cursor, elements=["Li", "P"], no_plot=True)
        hull.volume_curve()
        self.assertEqual(len(hull.volume_data["x"][0]), len(hull.hull_cursor) - 1)
        self.assertEqual(
            len(hull.volume_data["electrode_volume"][0]), len(hull.hull_cursor) - 1
        )

    def test_ternary_volume_curve(self):
        cursor = [
            {
                "stoichiometry": [["K", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 10,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["Sn", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["P", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["K", 1.0], ["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": -1,
                "cell_volume": 210,
                "num_atoms": 30,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["Sn", 2.0], ["P", 1.0]],
                "enthalpy_per_atom": -0.1,
                "cell_volume": 300,
                "num_atoms": 30,
                "num_fu": 10,
            },
        ]
        for ind, doc in enumerate(cursor):
            cursor[ind]["concentration"] = get_concentration(doc, ["K", "Sn", "P"])
            cursor[ind]["source"] = ["abcde"]
            cursor[ind]["enthalpy"] = (
                cursor[ind]["enthalpy_per_atom"] * doc["num_atoms"]
            )

        hull = QueryConvexHull(
            cursor=cursor, elements=["K", "Sn", "P"], no_plot=True, voltage=True
        )

        np.testing.assert_array_almost_equal(
            hull.volume_data["electrode_volume"][0], [30, 31]
        )
        np.testing.assert_array_almost_equal(
            hull.volume_data["volume_ratio_with_bulk"][0], [1, 31 / 30]
        )

        cursor = [
            {
                "stoichiometry": [["K", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 10,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["Sn", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["P", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["K", 1.0], ["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": -1,
                "cell_volume": 210,
                "num_atoms": 30,
                "num_fu": 10,
            },
            {
                "stoichiometry": [["Sn", 2.0], ["P", 1.0]],
                "enthalpy_per_atom": -0.1,
                "cell_volume": 150,
                "num_atoms": 30,
                "num_fu": 10,
            },
        ]
        for ind, doc in enumerate(cursor):
            cursor[ind]["concentration"] = get_concentration(doc, ["K", "Sn", "P"])
            cursor[ind]["source"] = ["abcde"]
            cursor[ind]["enthalpy"] = (
                cursor[ind]["enthalpy_per_atom"] * doc["num_atoms"]
            )

        hull = QueryConvexHull(
            cursor=cursor, elements=["K", "Sn", "P"], no_plot=True, voltage=True
        )

        np.testing.assert_array_almost_equal(
            hull.volume_data["electrode_volume"][0], [15, 31]
        )
        np.testing.assert_array_almost_equal(
            hull.volume_data["volume_ratio_with_bulk"][0], [1, 31 / 15]
        )

    def test_realistic_ternary(self):
        cursor, f = res2dict(REAL_PATH + "/data/voltage_data/voltage-LiCoP-oqmd/*.res")
        hull = QueryConvexHull(
            cursor=cursor, species="LiCoP", voltage=True, volume=True, no_plot=True
        )

        for key in hull.volume_data:
            self.assertEqual(len(hull.volume_data[key]), 4)

        expected = [
            [1, 1.6635, 2.6445, 3.1302],
            [1, 2.4353, 3.1457],
            [1, 1.0983, 1.7114, 2.9253],
            [1, 1.1339, 2.44096],
        ]
        for i in range(4):
            np.testing.assert_array_almost_equal(
                np.asarray(hull.volume_data["volume_ratio_with_bulk"][i]),
                expected[i],
                decimal=4,
            )

        expected_reactions = [
            [
                [(1.0, "CoP3")],
                [(1.0, "CoP2"), (1.0, "Li3P")],
                [(1 / 6, "Co6LiP4", 3 - 4 / 6, "Li3P")],
                [(3.0, "Li3P"), (1.0, "Co")],
            ],
            [
                [(1.0, "CoP2")],
                [(1 / 6, "Co6LiP4"), (2 - 4 / 6, "Li3P")],
                [(2.0, "Li3P"), (1.0, "Co")],
            ],
            [
                [(1.0, "CoP")],
                [(1 / 4, "CoP2"), (3 / 6 * 1 / 4, "Co6LiP4")],
                [(1 / 6, "Co6LiP4"), (1 - 4 / 6, "Li3P")],
                [(1, "Li3P"), (1, "Co")],
            ],
            [
                [(1.0, "Co2P")],
                [(1 / 4, "Co6LiP4"), (2 - 6 / 4, "Co")],
                [(1.0, "Li3P"), (2.0, "Co")],
            ],
        ]

        for i in range(len(expected_reactions)):
            for j, react in enumerate(expected_reactions[i]):
                for k, elem in enumerate(react):
                    self.assertEqual(
                        hull.voltage_data[i].reactions[j][k][1], elem[1]
                    )
                    self.assertAlmostEqual(
                        hull.voltage_data[i].reactions[j][k][0],
                        elem[0],
                        msg="\nReaction ({}, {}, {}) did not match: {} vs {}".format(
                            i, j, k, hull.voltage_data[i].reactions[j][k], elem
                        ),
                    )



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The calculator module contains classes for use as DFT/atomistic
calculators within the compute module class.

"""


__all__ = ['Calculator', 'CastepCalculator']  # , 'QuantumEspressoCalculator', 'ASECalculator']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.calculators.castep import CastepCalculator
# from matador.calculators.quantum_espresso import QuantumEspressoCalculator
# from matador.calculators.ase import ASECalculator



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule defines some useful chemical functions and
constants, with a focus on battery materials.

"""


import copy
import warnings

import numpy as np
from matador.data.constants import * # noqa

EPS = 1e-8


def get_iupac_ordering():
    """ Stub for implementing IUPAC chemical ordering in formulae. """
    raise NotImplementedError


def get_periodic_table():
    """ Return some periodic table macros. """
    periodic_table = dict()
    periodic_table['I'] = ['Li', 'Na', 'K', 'Rb', 'Cs', 'Fr']
    periodic_table['II'] = ['Be', 'Mg', 'Ca', 'Sr', 'Ba', 'Ra']
    periodic_table['III'] = ['B', 'Al', 'Ga', 'In', 'Tl']
    periodic_table['IV'] = ['C', 'Si', 'Ge', 'Sn', 'Pb']
    periodic_table['V'] = ['N', 'P', 'As', 'Sb', 'Bi']
    periodic_table['VI'] = ['O', 'S', 'Se', 'Te', 'Po']
    periodic_table['VII'] = ['F', 'Cl', 'Br', 'I', 'At']
    periodic_table['Tran'] = ['Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
                              'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
                              'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg']
    periodic_table['Lan'] = ['La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu']
    periodic_table['Act'] = ['Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']
    periodic_table['X'] = [elem for group in periodic_table for elem in periodic_table[group]]
    return periodic_table


def get_molar_mass(elem):
    """ Returns molar mass of chosen element. """
    import periodictable
    return periodictable.elements.symbol(elem).mass


def get_atomic_number(elem):
    """ Returns atomic number of chosen element. """
    import periodictable
    return periodictable.elements.symbol(elem).number


def get_atomic_symbol(atomic_number):
    """ Returns elemental symbol from atomic number. """
    import periodictable
    return periodictable.elements[atomic_number].symbol


def get_concentration(doc, elements, include_end=False):
    """ Returns x for A_x B_{1-x}
    or x,y for A_x B_y C_z, (x+y+z=1).

    Parameters:
        doc (list/dict): structure to evaluate OR matador-style stoichiometry.
        elements (list): list of element symbols to enforce ordering.

    Keyword arguments:
        include_end (bool): whether or not to return the final value, i.e.
            [x, y, z] rather than [x, y] in the above.

    Returns:
        list of float: concentrations of elements in given order.

    """
    if isinstance(doc, dict):
        if doc.get('stoichiometry') is None:
            raise RuntimeError('No stoichiometry found.')

        stoich = doc['stoichiometry']
    else:
        stoich = doc

    concs = [0.0] * (len(elements) - bool(not include_end))
    for _, elem in enumerate(stoich):
        if (include_end and elem[0] in elements) or (not include_end and elem[0] in elements[:-1]):
            concs[elements.index(elem[0])] = elem[1] / float(get_atoms_per_fu(doc))
    return concs


def get_num_intercalated(cursor):
    """ Return array of the number of intercalated atoms
    per host atom from a list of structures, of type defined by
    the first entry in the structures' concentration vectors.

    Parameters:
        cursor (list of dict): structures to evaluate.

    Returns:
        ndarray: number of intercalated ions in each structure.

    """
    from .cursor_utils import get_array_from_cursor
    x = np.zeros((len(cursor)))
    comps = get_array_from_cursor(cursor, 'concentration')
    for idx, comp in enumerate(comps):
        if len(comp) > 1:
            comp = comp[0]
        if 1 - comp == 0:
            x[idx] = np.NaN
        else:
            x[idx] = comp / (1 - comp)
    return x


def get_binary_grav_capacities(x, m_B):
    """ Returns capacity in mAh/g from x/y in A_x B_y
    and m_B in a.m.u.
    """
    x = np.array(x)
    if m_B != 0:
        return x * FARADAY_CONSTANT_Cpermol * Cperg_to_mAhperg / m_B
    return float('NaN')


def get_generic_grav_capacity(concs, elements):
    """ Returns gravimetric capacity of
    <elements[0]> in mAh/g of matador doc.
    """
    tmp_concs = np.array(concs, copy=True)
    # if no Li, capacity = 0...
    # tmp_concs /= np.min(concs)
    x = tmp_concs[0]
    if x == 0:
        return 0.0
    masses = dict()
    m_B = 0
    for elem in elements:
        masses[elem] = get_molar_mass(elem)
    for ind, elem in enumerate(elements):
        if ind == 0:
            continue
        else:
            m_B += masses[elem] * tmp_concs[ind]
    Q = get_binary_grav_capacities(x, m_B)
    return Q


def get_binary_volumetric_capacity(initial_doc, final_doc):
    """ For initial (delithiated/sodiated) (single element) structure
    and final (maximally charged) binary structure, calculate the volumetric capacity.

    Parameters:
        initial_doc (dict): matador doc of delithiated phase
        final_doc (dict): matador doc of maximally lithiated phase

    Returns:
       volumetric_capacity (float): capacity in mAh/cm^3.

    """

    assert len(initial_doc['stoichiometry']) == 1
    assert len(final_doc['stoichiometry']) == 2

    for _ion in final_doc['stoichiometry']:
        if _ion[0] != initial_doc['stoichiometry'][0][0]:
            ion = _ion[0]

    for species in final_doc['stoichiometry']:
        if species[0] == ion:
            num_ion = species[1]
        else:
            num_B = species[1]

    num_ions_per_initial_fu = num_ion / num_B
    volume_per_fu_cm3 = initial_doc['cell_volume'] * ANGSTROM_CUBED_TO_CENTIMETRE_CUBED / initial_doc['num_fu']
    return (num_ions_per_initial_fu / volume_per_fu_cm3) * (ELECTRON_CHARGE * Cperg_to_mAhperg)


def get_atoms_per_fu(doc):
    """ Calculate and return the number of atoms per formula unit.

    Parameters:
        doc (list/dict): structure to evaluate OR matador-style stoichiometry.

    """
    if 'stoichiometry' in doc:
        return sum([elem[1] for elem in doc['stoichiometry']])
    return sum([elem[1] for elem in doc])


def get_formation_energy(chempots, doc, energy_key='enthalpy_per_atom'):
    """ From given chemical potentials, calculate the simplest
    formation energy per atom of the desired document.

    Note:
        recursive_get(doc, energy_key) MUST return an energy per atom for
        the target doc and the chemical potentials.

    Parameters:
        chempots (list of dict): list of chempot structures, must be unique.
        doc (dict): structure to evaluate.

    Keyword arguments:
        energy_key (str or list): name of energy field to use to calculate
            formation energy. Can use a list of keys/subkeys/indices to
            query nested dicts with `matador.utils.cursor_utils.recursive_get`.

    Returns:
        float: formation energy per atom.

    """
    from matador.utils.cursor_utils import recursive_get

    # warn user if per_atom energy is not found
    if isinstance(energy_key, (list, tuple)) and not any(['per_atom' in str(key) for key in energy_key]) \
            or (isinstance(energy_key, str) and 'per_atom' not in energy_key):
        warnings.warn('Requested energy key {} in get_formation_energy may'
                      ' not be per atom, if so results will be incorrect.'.format(energy_key))

    try:
        formation = recursive_get(doc, energy_key)
    except KeyError as exc:
        print('Doc {} missing key {}'.format(doc['source'], energy_key))
        raise exc

    # see if num chempots has been set and try to reuse it
    if 'num_chempots' in doc:
        num_chempots = doc['num_chempots']
    else:
        num_chempots = get_number_of_chempots(doc, chempots)

    num_atoms_per_fu = get_atoms_per_fu(doc)
    for ind, mu in enumerate(chempots):
        num_atoms_per_mu = get_atoms_per_fu(mu)
        try:
            mu_energy = recursive_get(mu, energy_key)
        except KeyError as exc:
            raise exc('Chemical potential {} missing key {}'.format(mu['source'], energy_key))

        formation -= mu_energy * num_chempots[ind] * num_atoms_per_mu / num_atoms_per_fu

    return formation


def get_number_of_chempots(stoich, chempot_stoichs, precision=5):
    """ Return the required number of each (arbitrary) chemical potentials
    to construct one formula unit of the input stoichiometry. Uses least-squares
    as implemented by `numpy.linalg.lstsq` and rounds the output precision based
    on the `precision` kwarg.

    Parameters:
        stoich (list/dict): matador-style stoichiometry,
            e.g. [['Li', 3], ['P', 1]], or the full document.
        chempot_stoichs (list/dict): list of stoichiometries of the input
            chemical potentials, or the full documents.

    Keyword arguments:
        precision (int/None): number of decimal places to round answer to. None
            maintains the precision from `numpy.linalg.lstsq`.

    Returns:
        list: number of each chemical potential required to create
            1 formula unit.

    Raises:
        RuntimeError: if the stoichiometry provided cannot be created
            with the given chemical potentials.

    """

    # need to support dict, list and Crystal inputs
    try:
        stoich = stoich['stoichiometry']
    except (TypeError, ValueError):
        pass
    try:
        chempot_stoichs = [mu['stoichiometry'] for mu in chempot_stoichs]
    except (TypeError, ValueError):
        pass

    # find all elements present in the chemical potentials
    elements = set()
    for mu in chempot_stoichs:
        for elem, _ in mu:
            elements.add(elem)
    elements = sorted(list(elements))

    chempot_matrix = np.asarray([get_padded_composition(mu, elements) for mu in chempot_stoichs])

    try:
        solution = np.asarray(get_padded_composition(stoich, elements))
    except RuntimeError:
        raise RuntimeError('Stoichiometry {} could not be created from chemical potentials {}: missing chempot'
                           .format(stoich, chempot_stoichs))

    try:
        num_chempots, residuals, _, _ = np.linalg.lstsq(chempot_matrix.T, solution, rcond=None)
    except np.linalg.LinAlgError:
        raise RuntimeError('Stoichiometry {} could not be created from chemical potentials {}: numpy LinAlg error'
                           .format(stoich, chempot_stoichs))

    # check if lstsq actually found a "solution"
    if np.abs(np.sum(residuals)) > EPS:
        raise RuntimeError('Stoichiometry {} could not be created from chemical potentials {}'
                           .format(stoich, chempot_stoichs))

    # round output array based on user-specified precision
    if precision is not None:
        num_chempots[np.where(np.abs(num_chempots) < EPS)] = 0.0
        for i, val in enumerate(num_chempots):
            if np.abs(val - round(val, precision)) < EPS:
                num_chempots[i] = round(val, precision)

    # check for sensible numbers in output
    if np.min(np.sign(num_chempots)) == -1:
        raise RuntimeError('Stoichiometry {} could not be created from chemical potentials {}'
                           .format(stoich, chempot_stoichs))

    return num_chempots.tolist()


def get_stoich(atom_types):
    """ Return integer stoichiometry from atom_types list.

    Parameters:
        atom_types (list): list of element symbols of each atom.

    Returns:
        list: matador-style stoichiometry, e.g. [['Li', 1], ['P', 2]].

    """
    from collections import defaultdict
    from math import gcd
    stoich = defaultdict(float)
    for atom in atom_types:
        if atom not in stoich:
            stoich[atom] = 0
        stoich[atom] += 1
    gcd_val = 0
    for atom in atom_types:
        if gcd_val == 0:
            gcd_val = stoich[atom]
        else:
            gcd_val = gcd(stoich[atom], gcd_val)
    # convert stoichiometry to tuple for fryan
    temp_stoich = []
    try:
        for key, value in stoich.items():
            if float(value) / gcd_val % 1 != 0:
                temp_stoich.append([key, float(value) / gcd_val])
            else:
                temp_stoich.append([key, value / gcd_val])
    except AttributeError:
        for key, value in stoich.iteritems():
            if float(value) / gcd_val % 1 != 0:
                temp_stoich.append([key, float(value) / gcd_val])
            else:
                temp_stoich.append([key, value / gcd_val])
    return sorted(temp_stoich)


def get_padded_composition(stoichiometry, elements):
    """ Return a list that contains how many of each species in
    elements exists in the given stoichiometry. e.g. for [['Li', 2], ['O', 1]]
    with elements ['O', 'Li', 'Ba'], this function will return [1, 2, 0].

    Parameters:
        stoichiometry (list): matador-style stoichiometry, as above.
        elements (list): order of element labels to pick out.

    """
    composition = []
    for element in elements:
        if not isinstance(element, str):
            raise RuntimeError('Found invalid element symbol {}'.format(element))
        for species in stoichiometry:
            if not isinstance(species, list):
                raise RuntimeError('Found invalid stoichiometry {}'.format(stoichiometry))
            if species[0] == element:
                composition.append(species[1])
                break
            elif species[0] not in elements:
                raise RuntimeError('Extra element {} in stoichiometry'.format(species[0]))
        else:
            composition.append(0)

    return composition


def get_ratios_from_stoichiometry(stoichiometry):
    """ Get a dictionary of pairwise atomic ratios.

    Parameters:
        stoichiometry (list): matador-style stoichiometry.

    Returns:
        dict: dictionary of pairwise ratios, e.g. for K8SnP4,
            ratio_dict = {'KSn': 8, 'KP': 2, 'SnP': 0.25,
                          'SnK': 0.125, 'PSn': 4, 'PK': 0.5}.

    """
    ratio_dict = dict()
    for _, elem_i in enumerate(stoichiometry):
        for _, elem_j in enumerate(stoichiometry):
            if elem_j[0] != elem_i[0]:
                ratio_dict[elem_i[0]+elem_j[0]] = round(float(elem_i[1]) / elem_j[1], 3)
    return ratio_dict


def get_stoich_from_formula(formula: str, sort=True):
    """ Convert formula string, e.g. Li2TiP4 into a matador-style
    stoichiometry, e.g. [['Li', 2], ['Ti', 1], ['P', 4]].

    Parameters:
        formula (str): chemical formula of compound

    Returns:
        list: sorted matador-style stoichiometry.

    """
    from math import gcd
    import re

    parsed_elements = parse_element_string(formula, stoich=True)
    elements = []
    fraction = []
    for i, _ in enumerate(parsed_elements):
        if not bool(re.search(r'\d', parsed_elements[i])):
            elements.append(parsed_elements[i])
            try:
                fraction.append(float(parsed_elements[i + 1]))
            except(ValueError, IndexError):
                fraction.append(1.0)
    gcd_val = 0
    for frac in fraction:
        if gcd_val == 0:
            gcd_val = frac
        else:
            gcd_val = gcd(int(frac), int(gcd_val))
    fraction = np.asarray(fraction)
    fraction /= gcd_val
    stoich = [[elements[ind], fraction[ind]] for ind, _ in enumerate(elements)]
    if sort:
        return sorted(stoich)

    return stoich


def parse_element_string(elements_str, stoich=False):
    """ Parse element query string with macros. Has to parse braces
    too, and throw an error if brackets are unmatched.

    e.g.
        Parameters: '[VII][Fe,Ru,Os][I]'
        Returns: ['[VII]', '[Fe,Ru,Os]', '[I]']

    e.g.2
        Parameters: '[VII]2[Fe,Ru,Os][I]'
        Returns: ['[VII]2', '[Fe,Ru,Os]', '[I]']

    Parameters:
        elements_str: str, chemical formula, including macros.

    Keyword arguments:
        stoich: bool, parse as a stoichiometry, i.e. check for numbers

    Raises:
        RuntimeError: if the composition contains unmatched brackets.

    Returns:
        list: split list of elements contained in input

    """
    import re
    valid = False
    for char in elements_str:
        if char not in ['[', ']', '{', '}', ',', ':'] and not char.isalnum():
            raise RuntimeError('Illegal character {} detected in query.'.format(char))
    valid = False
    for char in elements_str:
        if char.isupper():
            valid = True
            break
    if not valid:
        raise RuntimeError('Composition must contain at least one upper case character.')
    elements = [elem for elem in re.split(r'([A-Z][a-z]*)', elements_str) if elem]
    if stoich:
        tmp_stoich = elements
        for ind, strng in enumerate(elements):
            if not any(char.isdigit() for char in strng):
                tmp_stoich[ind] = [strng]
            else:
                tmp_stoich[ind] = [elem for elem in re.split(r'([0-9]+)', strng) if elem]
        elements = [item for sublist in tmp_stoich for item in sublist]
    # split macros
    while '[' in elements or '{' in elements or '][' in elements or '}{' in elements or ']{' in elements or '}[' in elements:
        tmp_stoich = list(elements)
        cleaned = True
        while cleaned:
            for ind, tmp in enumerate(tmp_stoich):
                if tmp == '][':
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, '[')
                    tmp_stoich.insert(ind, ']')
                    cleaned = True
                elif tmp == '}{':
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, '{')
                    tmp_stoich.insert(ind, '}')
                    cleaned = True
                elif tmp == ']{':
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, '{')
                    tmp_stoich.insert(ind, ']')
                    cleaned = True
                elif tmp == '}[':
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, '[')
                    tmp_stoich.insert(ind, '}')
                    cleaned = True
                elif ind == len(tmp_stoich) - 1:
                    cleaned = False
        for ind, tmp in enumerate(tmp_stoich):
            if tmp == '[':
                end_bracket = False
                while not end_bracket:
                    if tmp_stoich[ind + 1] == ']':
                        end_bracket = True
                    tmp_stoich[ind] += tmp_stoich[ind + 1]
                    del tmp_stoich[ind + 1]
        for ind, tmp in enumerate(tmp_stoich):
            if tmp == '{':
                end_bracket = False
                while not end_bracket:
                    if tmp_stoich[ind + 1] == '}':
                        end_bracket = True
                    tmp_stoich[ind] += tmp_stoich[ind + 1]
                    del tmp_stoich[ind + 1]
        if ']' in tmp_stoich:
            tmp_stoich.remove(']')
        if '}' in tmp_stoich:
            tmp_stoich.remove('}')
        if '' in tmp_stoich:
            tmp_stoich.remove('')

        elements = tmp_stoich

    for elem in elements:
        if '}[' in elem or ']{' in elem:
            raise RuntimeError('Unmatched brackets in query string')

    return elements


def get_root_source(source):
    """ Get the main file source from a doc's source list.

    Parameters:
        source (list/dict): contents of doc['source'] or the doc itself.

    Returns:
        str: "root" filename, e.g. if source = ['KP.cell', 'KP.param',
            'KP_specific_structure.res'] then root = 'KP_specific_structure'.

    """
    try:
        sources = copy.deepcopy(source['source'])
    except (KeyError, TypeError):
        sources = copy.deepcopy(source)

    src_list = set()
    for src in sources:
        if any([src.endswith(ext) for ext in
               ['.res', '.castep', '.history', '.history.gz', '.phonon', '.phonon_dos', '.bands', '.cif']]):
            src_list.add('.'.join(src.split('/')[-1].split('.')[0:-1]))
        elif 'OQMD' in src.upper():
            src_list.add(src)
        elif 'MP-' in src.upper():
            src_list.add(src)
        elif len(sources) == 1:
            src_list.add(src)
        elif src == 'command_line':
            src_list.add('command line')

    if len(src_list) > 1:
        raise RuntimeError('Ambiguous root source {}'.format(sources))
    if len(src_list) < 1:
        raise RuntimeError('Unable to find root source from {}'.format(sources))

    return list(src_list)[0]


def get_formula_from_stoich(stoich, elements=None, tex=False, sort=True, latex_sub_style=''):
    """ Get the chemical formula of a structure from
    its matador stoichiometry.

    Parameters:
        stoich (list): matador-style stoichiometry.

    Keyword arguments:
        elements (list): list of element symbols to enforce order.
        tex (bool): whether to print a LaTeX-compatibile string.
        latex_sub_style (str): a string to wrap subscripts in, e.g.
            r"\\mathrm" or r"\\text" (default is blank).

    Returns:
        str: the string representation of the chemical formula.

    """
    form = ''
    if not isinstance(stoich, list):
        stoich = stoich.tolist()
    if sort:
        stoich = sorted(stoich)
    if elements is not None:
        for targ_elem in elements:
            for elem in stoich:
                if elem[0] == targ_elem:
                    if elem[1] == 1:
                        form += elem[0]
                    elif int(elem[1]) != 0:
                        if tex:
                            if elem[1] % 1 == 0:
                                elem[1] = int(elem[1])
                            form += r'{}$_{}{{{}}}$'.format(elem[0], latex_sub_style, elem[1])
                        else:
                            form += elem[0] + str(int(elem[1]))
        assert form != ''
    else:
        for elem in stoich:
            if elem[1] == 1:
                form += elem[0]
            elif int(elem[1]) != 0:
                if tex:
                    if elem[1] % 1 == 0:
                        elem[1] = int(elem[1])
                    form += '{}$_{}{{{}}}$'.format(elem[0], latex_sub_style, elem[1])
                else:
                    form += elem[0] + str(int(elem[1]))
    return form



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some scraper functions for
NMR-related inputs and outputs, e.g. .magres files.

"""


from collections import defaultdict
from os import stat
from pwd import getpwuid

import numpy as np
from matador.utils.cell_utils import cart2abc, cart2frac
from matador.utils.chem_utils import get_stoich
from matador.scrapers.utils import scraper_function


@scraper_function
def magres2dict(seed, **kwargs):
    """ Extract available information from .magres file. Assumes units of
    Angstrom and ppm for relevant quantities.
    """
    magres = defaultdict(list)
    # read .pwout file into array
    seed = seed.replace('.magres', '')
    with open(seed + '.magres', 'r') as f:
        flines = f.readlines()
    # add .magres to source
    magres['source'].append(seed+'.magres')
    # grab file owner username
    try:
        magres['user'] = getpwuid(stat(seed + '.magres').st_uid).pw_name
    except Exception:
        magres['user'] = 'xxx'
        if kwargs.get('debug'):
            print(seed + '.magres has no owner.')

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<atoms>', '[atoms]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</atoms>', '[/atoms]']:
                if i > 1e5:
                    raise RuntimeError
                if 'units' in flines[line_no+i]:
                    i += 1
                    continue
                elif 'lattice' in flines[line_no+i]:
                    lattice = flines[line_no+i].split()[1:]
                    for j in range(3):
                        magres['lattice_cart'].append([float(elem) for elem in lattice[j*3:(j+1)*3]])
                    magres['lattice_abc'] = cart2abc(magres['lattice_cart'])
                elif 'atom' in flines[line_no+i]:
                    atom = flines[line_no+i].split()
                    magres['atom_types'].append(atom[1])
                    magres['positions_abs'].append([float(elem) for elem in atom[-3:]])
                i += 1
            break

    magres['num_atoms'] = len(magres['atom_types'])
    magres['positions_frac'] = cart2frac(magres['lattice_cart'], magres['positions_abs'])
    magres['stoichiometry'] = get_stoich(magres['atom_types'])

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<magres>', '[magres]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</magres>', '[/magres]']:
                if i > 1e5:
                    raise RuntimeError
                if 'units' in flines[line_no+i]:
                    i += 1
                    continue
                elif 'sus' in flines[line_no+i]:
                    sus = flines[line_no+i].split()[1:]
                    for j in range(3):
                        magres['susceptibility_tensor'].append([float(val) for val in sus[3*j:3*(j+1)]])
                elif 'ms' in flines[line_no+i]:
                    ms = flines[line_no+i].split()[3:]
                    magres['magnetic_shielding_tensor'].append([])
                    for j in range(3):
                        magres['magnetic_shielding_tensor'][-1].append([float(val) for val in ms[3*j:3*(j+1)]])
                    magres['chemical_shifts'].append(0)
                    magres['chemical_shift_anisos'].append(0)
                    magres['chemical_shift_asymmetries'].append(0)
                    for j in range(3):
                        magres['chemical_shifts'][-1] += magres['magnetic_shielding_tensor'][-1][j][j] / 3

                    # find eigenvalues of symmetric part of shielding and order them to calc anisotropy eta
                    symmetric_shielding = (
                        0.5 *
                        (magres['magnetic_shielding_tensor'][-1] + np.asarray(magres['magnetic_shielding_tensor'][-1]).T)
                    )
                    eig_vals, eig_vecs = np.linalg.eig(symmetric_shielding)
                    eig_vals, eig_vecs = zip(*sorted(zip(eig_vals, eig_vecs),
                                                     key=lambda eig: abs(eig[0] - magres['chemical_shifts'][-1])))
                    # Haeberlen convention: |s_zz - s_iso| >= |s_xx - s_iso| >= |s_yy - s_iso|
                    s_yy, s_xx, s_zz = eig_vals
                    s_iso = magres['chemical_shifts'][-1]
                    # convert from reduced anistropy to CSA
                    magres['chemical_shift_anisos'][-1] = s_zz - (s_xx + s_yy)/2.0
                    magres['chemical_shift_asymmetries'][-1] = (s_yy - s_xx) / (s_zz - s_iso)
                i += 1

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ['<calculation>', '[calculation]']:
            i = 1
            while flines[line_no+i].strip().lower() not in ['</calculation>', '[/calculation]']:
                if i > 1e5:
                    raise RuntimeError
                # space important as it excludes other calc_code_x variables
                if 'calc_code ' in flines[line_no+i]:
                    magres['calculator'] = flines[line_no+i].split()[1]
                if 'calc_code_version' in flines[line_no+i]:
                    magres['calculator_version'] = flines[line_no+i].split()[1]
                i += 1

    return magres, True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements a simple interface to basic SLURM
functionality, including creating and submitting slurm scripts and
cancelling jobs.

"""

from matador.compute.queueing import QueueManager


class SlurmQueueManager(QueueManager):
    """ Wrapper for the Slurm queueing system. """

    token = 'slurm'

    def get_array_id(self):
        if self.env.get('SLURM_ARRAY_TASK_ID') is not None:
            return int(self.env['SLURM_ARRAY_TASK_ID'])
        return None

    def get_ntasks(self):
        return int(self.env['SLURM_NTASKS'])

    def get_max_memory(self):
        if self.env.get('SLURM_MEM_PER_CPU') is None:
            return None
        else:
            return float(self.env['SLURM_MEM_PER_CPU']) * self.ntasks

    def get_walltime(self):
        """ Query available walltime with scontrol on the current job.

        Parameters:
            slurm_dict (dict): slurm env parameters to query.

        Raises:
            RuntimeError: if SLURM_JOB_ID not present in slurm env.
            subprocess.CalledProcessError: if unable to use scontrol.

        Returns:
            int: maximum allowed walltime time in seconds.

        """
        import subprocess as sp
        slurm_dict = self.env
        job_id = slurm_dict.get('SLURM_JOB_ID')
        if job_id is None:
            return None

        output = sp.check_output('scontrol show job={}'.format(job_id), shell=True).decode('utf-8')
        output_dict = {line.split('=')[0].lower(): line.split('=')[-1] for line in output.split()}

        walltime = output_dict.get('timelimit')
        hrs = 0
        if '-' in walltime:
            days = int(walltime.split('-')[0])
            walltime = walltime.split('-')[1]
            hrs += days * 24

        hrs += int(walltime.split(':')[0])
        mins = int(walltime.split(':')[1])
        secs = int(walltime.split(':')[2])
        walltime_in_seconds = (60 * hrs + mins) * 60 + secs

        return walltime_in_seconds


def scancel_all_matching_jobs(name=None):
    """ Cancel all of the user's jobs.

    Keyword arguments:
        name (str): optional name to pass to scancel

    Returns:
        str: output from scancel.

    """
    from os import getlogin
    import subprocess as sp
    user = getlogin()
    if name is None:
        return sp.check_output('scancel -u {}'.format(user), shell=True).decode('utf-8')

    return sp.check_output('scancel -u {} -n {}'.format(user, name), shell=True).decode('utf-8')


def submit_slurm_script(slurm_fname, depend_on_job=None, num_array_tasks=None):
    """ Submit a SLURM job.

    Parameters:
        slurm_fname (str): SLURM job file to submit.

    Keyword arguments:
        depend_on_job (int): job ID to make current job depend on.
        num_array_tasks (int): number of array tasks to submit.

    Raises:
        subprocess.CalledProcessError: if jobfile doesn't exist or has failed.

    Return:
        int: submitted SLURM job ID.

    """
    import subprocess as sp
    command = 'sbatch '
    if depend_on_job is not None:
        command += '--dependency=afterany:{} '.format(depend_on_job)
    if num_array_tasks is not None:
        assert num_array_tasks > 0
        if num_array_tasks != 1:
            command += '--array=0-{} '.format(num_array_tasks-1)
    command += '{}'.format(slurm_fname)
    slurm_output = sp.check_output(command, shell=True).decode('utf-8')
    slurm_job_id = int(slurm_output.strip().split()[-1])
    return slurm_job_id


def get_slurm_header(slurm_dict, walltime_hrs, num_nodes=None):
    """ Write a SLURM script header from a set of slurm parameters.

    Parameters:
        slurm_dict (dict): dictionary of SLURM environment variables.
        walltime_hrs (int): allowed walltime in hours

    Keyword arguments:
        num_nodes (int): overrides $SLURM_JOB_NUM_NODES with a custom value.

    Returns:
        header (str): the SLURM file header.

    """

    header = "#!/bin/bash\n"
    header += "#! SLURM file written by matador (Matthew Evans 2016-2018).\n\n"
    header += "#! Name of job:\n"
    header += "#SBATCH --job-name {}\n".format(slurm_dict['SLURM_JOB_NAME'])
    header += "#! Name of project:\n"
    header += "#SBATCH --account {}\n".format(slurm_dict['SLURM_JOB_ACCOUNT'])
    if num_nodes is None:
        num_nodes = slurm_dict['SLURM_JOB_NUM_NODES']
    header += "#! Number of nodes to allocate:\n"
    header += "#SBATCH --nodes {}\n".format(num_nodes)
    header += "#! Number of tasks to allocate:\n"
    header += "#SBATCH --ntasks {}\n".format(slurm_dict['SLURM_NTASKS'])
    header += "#! Partition:\n"
    header += "#SBATCH --partition {}\n".format(slurm_dict['SLURM_JOB_PARTITION'])
    header += "#! Walltime to allocate:\n"
    header += "#SBATCH --time {}:00:00\n".format(walltime_hrs)

    return header


def write_slurm_submission_script(slurm_fname, slurm_dict, compute_string, walltime_hrs, template=None):
    """ Write a full slurm submission script based on the
    input settings.

    Parameters:
        slurm_fname (str): the desired filename for the submission script
        slurm_dict (dict): dictionary of SLURM environment variables
        compute_string (str): the compute commands to run
        walltime_hrs (int): maximum walltime in hours

    Keyword arguments:
        template (str): filename containing job preamble, e.g. module loads

    """
    header = get_slurm_header(slurm_dict, walltime_hrs)
    if template is not None:
        with open(template, 'r') as f:
            preamble = f.readlines()
    else:
        preamble = []

    with open(slurm_fname, 'w') as f:
        f.write(header)
        f.write('\n\n')
        for line in preamble:
            f.write(line)
        f.write('\n\n')
        f.write(compute_string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# matador documentation build configuration file, created by
# sphinx-quickstart on Tue Jan  3 17:43:56 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sphinx_rtd_theme

from matador import __version__

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.napoleon",
    "sphinxarg.ext",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "nbsphinx",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
]
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

autodoc_member_order = "bysource"
autoclass_content = "both"

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "matador"
copyright = "2016-2020, Matthew Evans"
author = "Matthew Evans"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# run apidoc automatically on RTD: https://github.com/rtfd/readthedocs.org/issues/1139
def run_apidoc(_):
    import subprocess
    import glob
    import shutil

    try:
        if not os.path.isfile("docs/src/img/lipzn.png"):
            os.makedirs("docs/src/img")
            shutil.copy("img/lipzn.png", "docs/src/img/lipzn.png")
    except Exception:
        pass
    src_dir = os.path.abspath(os.path.dirname(__file__))
    excludes = glob.glob(os.path.join(src_dir, "../../matador/tests/"))
    module = os.path.join(src_dir, "../../matador")
    cmd_path = "sphinx-apidoc"
    print(excludes)
    command = [cmd_path, "-M", "-o", src_dir, module, " ".join(excludes)]
    print(command)
    subprocess.check_call(command)


def setup(app):
    app.connect("builder-inited", run_apidoc)


# -- Napolean options -----------------------------------------------------

# Napoleon settings
#
# napoleon_google_docstring = True
# napoleon_numpy_docstring = True
# napoleon_include_init_with_doc = False
# napoleon_include_private_with_doc = False
# napoleon_include_special_with_doc = False
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = False
# napoleon_use_param = False
# napoleon_use_rtype = False
# napoleon_use_keyword = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
# html_theme = 'alabaster'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "canonical_url": "https://docs.matador.science",
    "display_version": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "matadordoc"


# -- Options for LaTeX output ---------------------------------------------
latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "matador.tex", "matador Documentation", "Matthew Evans", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "matador", "matador Documentation", [author], 1)]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.6", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "pymongo": ("https://api.mongodb.com/python/current/", None),
    "np": ("http://docs.scipy.org/doc/numpy/", None),
    "matplotlib": ("http://matplotlib.org", None),
}



# coding: utf-8
# Distributed under the terms of the MIT License.

""" `matador` performs, aggregates, and analyses high-throughput
electronic structure calculations, primarily for crystal structure
prediction, leveraging CASTEP and Quantum Epsresso as density-functional
theory compute engines.
"""


__all__ = ['__version__']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'
__version__ = "0.9.3"

script_epilog = f"Written and maintained by Matthew Evans (me388@cam.ac.uk) 2016-2020, version {__version__}."



# coding: utf-8
# Distributed under the terms of the MIT License.

import os
from setuptools import setup, find_packages
from glob import glob

from matador import __version__

with open('requirements/requirements.txt', 'r') as f:
    requirements = [line.strip() for line in f.readlines()]

with open('requirements/pip_requirements.txt', 'r') as f:
    requirements += [line.strip() for line in f.readlines()]

extra_requirements = dict(all=[])
req_files = glob('requirements/*.txt')
for _file in req_files:
    if _file not in ['requirements/requirements.txt', 'requirements/pip_requirements.txt']:
        with open(_file, 'r') as f:
            subreq = _file.split('/')[-1].split('_')[0]
            extra_requirements[subreq] = [line.strip() for line in f.readlines()]
            extra_requirements['all'] += extra_requirements[subreq]

with open("README.rst", "r") as f:
    long_description = f.read()


setup(name='matador-db',
      version=__version__,
      description='MATerial and Atomic Databases Of Refined structures.',
      long_description=long_description,
      url='https://github.com/ml-evs/matador',
      author='Matthew Evans',
      author_email='matador@ml-evs.science',
      maintainer='Matthew Evans',
      maintainer_email='matador@ml-evs.science',
      license='MIT',
      packages=find_packages(),
      python_requires='>=3.6',
      install_requires=requirements,
      scripts=[script for script in glob('scripts/*') if os.path.isfile(script)],
      test_suite='matador.tests',
      include_package_data=True,
      extras_require=extra_requirements,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Chemistry",
          "Topic :: Scientific/Engineering :: Physics"
      ],
      entry_points={'console_scripts': ['matador = matador.cli.cli:main',
                                        'dispersion = matador.cli.dispersion:main',
                                        'run3 = matador.cli.run3:main']},
      zip_safe=False)



#!/usr/bin/env python

import unittest

try:
    import ase  # noqa

    ASE_IMPORTED = True
except ImportError:
    ASE_IMPORTED = False

try:
    import pymatgen  # noqa

    PMG_IMPORTED = True
except ImportError:
    PMG_IMPORTED = False


@unittest.skipIf(~(PMG_IMPORTED and ASE_IMPORTED), "Unable to import ASE/pymtagen")
class PMGUtilTest(unittest.TestCase):
    """ Tests cursor util functions. """

    def test_pm2dict(self):
        pass


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule defines some useful generic cursor methods for
displaying, extracting and refining results from a Mongo cursor/list.

"""

from time import strftime

import numpy as np
import pymongo as pm
from matador.utils.chem_utils import get_formula_from_stoich, get_root_source, get_stoich_from_formula
from matador import __version__

EPS = 1e-12


def recursive_get(data, keys, _top=True):
    """ Recursively slice a nested dictionary by a
    list of keys.

    Parameters:
        data (dict): nested dictionary to get from.
        keys (list): list of keys/indices to delve into.

    Raises:
        KeyError: if any in chain keys are missing,
        IndexError: if any element of a sublist is
            missing.

    """
    if not isinstance(keys, (list, tuple)):
        return data[keys]

    if isinstance(keys, (list, tuple)) and len(keys) == 1:
        return data[keys[0]]
    try:
        return recursive_get(data[keys[0]], keys[1:], _top=False)
    except (KeyError, IndexError) as exc:
        if _top:
            raise type(exc)('Recursive keys {} missing'.format(keys))
        raise exc


def recursive_set(data, keys, value):
    """ Recursively slice a nested dictionary by a
    list of keys and set the value.

    Parameters:
        data (dict): nested dictionary to get from.
        keys (list): list of keys/indices to delve into.
        value: value to store under key.

    Raises:
        KeyError: if any intermediate keys are missing.

    """
    if isinstance(keys, (list, tuple)):
        if len(keys) == 1:
            data[keys[0]] = value
        else:
            return recursive_set(data[keys[0]], keys[1:], value)
    else:
        data[keys] = value


def display_results(cursor,
                    energy_key='enthalpy_per_atom', summary=False,
                    args=None, argstr=None, additions=None, deletions=None,
                    sort=True, hull=False, markdown=False, latex=False, colour=True,
                    return_str=False, use_source=True, details=False,
                    per_atom=False, eform=False, source=False, **kwargs):
    """ Print query results in a table, with many options for customisability.

    TODO: this function has gotten out of control and should be rewritten.

    Parameters:
        cursor (list of dict or pm.cursor.Cursor): list of matador documents

    Keyword arguments:
        summary (bool): print a summary per stoichiometry, that uses the lowest
            energy phase (requires `sort=True`).
        argstr (str): string to store matador initialisation command
        eform (bool): prepend energy key with "formation_".
        sort (bool): sort input cursor by the value of energy key.
        return_str (bool): return string instead of printing.
        details (bool): print extra details as an extra line per structure.
        per_atom (bool): print quantities per atom, rather than per fu.
        source (bool): print all source files associated with the structure.
        use_source (bool): use the source instead of the text id when displaying a structure.
        hull (bool): whether or not to print hull-style (True) or query-style
        energy_key (str or list): key (or recursive key) to print as energy (per atom)
        markdown (bool): whether or not to write a markdown file containing results
        latex (bool): whether or not to create a LaTeX table
        colour (bool): colour on-hull structures
        additions (list): list of string text_ids to be coloured green with a (+)
            or, list of indices referring to those structures in the cursor.
        deletions (list): list of string text_ids to be coloured red with a (-)
            or, list of indices referring to those structures in the cursor.
        kwargs (dict): any extra args are ignored.

    Returns:
        str or None: markdown or latex string, if markdown or latex is True, else None.

    """

    add_index_mode = False
    del_index_mode = False
    if additions:
        if isinstance(additions[0], int):
            add_index_mode = True
    if deletions:
        if isinstance(deletions[0], int):
            del_index_mode = True

    if add_index_mode:
        assert max(additions) <= len(cursor) and min(additions) >= 0
    if del_index_mode:
        assert max(deletions) <= len(cursor) and min(deletions) >= 0

    if markdown and latex:
        latex = False

    # lists in which to accumulate the table
    struct_strings = []
    detail_strings = []
    detail_substrings = []
    source_strings = []
    formulae = []

    # tracking the last formula
    last_formula = ''

    if not cursor:
        raise RuntimeError('No structures found in cursor.')

    if markdown:
        markdown_string = 'Date: {}  \n'.format(strftime("%H:%M %d/%m/%Y"))
        if argstr is not None:
            markdown_string += 'Command: matador {}  \n'.format(' '.join(argstr))
        markdown_string += 'Version: {}  \n\n'.format(__version__)

    if latex:
        latex_string = (
            "\\begin{tabular}{l r r c l l}\n"
            "\\rowcolor{gray!20}\n"
            "formula & "
            "\\thead{$\\Delta E$\\\\(meV/atom)} & "
            "spacegroup & "
            "provenance & "
            "description \\\\ \n\n"
        )
        latex_struct_strings = []
        latex_sub_style = r'\mathrm'
    else:
        latex_sub_style = ''

    header_string, units_string = _construct_header_string(
        markdown, use_source, per_atom, eform, hull, summary, energy_key
    )

    if summary and isinstance(cursor, pm.cursor.Cursor):
        raise RuntimeError("Unable to provide summary when displaying cursor object.")

    # ensure cursor is sorted by enthalpy
    if sort and isinstance(cursor, pm.cursor.Cursor):
        print("Unable to check sorting of cursor, assuming it is already sorted.")
    elif sort:
        sorted_inds = sorted(enumerate(cursor),
                             key=lambda element: recursive_get(element[1], energy_key))
        cursor = [ind[1] for ind in sorted_inds]
        sorted_inds = [ind[0] for ind in sorted_inds]
        if additions is not None and add_index_mode:
            additions = [sorted_inds.index(ind) for ind in additions]
        if deletions is not None and del_index_mode:
            deletions = [sorted_inds.index(ind) for ind in deletions]

    # loop over structures and create pretty output
    for ind, doc in enumerate(cursor):

        # use the formula to see if we need to update gs_enthalpy for this formula
        formula_substring = get_formula_from_stoich(doc['stoichiometry'],
                                                    tex=latex,
                                                    latex_sub_style=latex_sub_style)

        if 'encapsulated' in doc:
            formula_substring += '+CNT'
        if last_formula != formula_substring:
            gs_enthalpy = 0.0
        formulae.append(formula_substring)

        struct_strings.append(
            _construct_structure_string(
                doc, ind, formula_substring, gs_enthalpy, use_source, colour, hull, additions,
                deletions, add_index_mode, del_index_mode, energy_key, per_atom, eform, markdown, latex
            )
        )

        if latex:
            latex_struct_strings.append("{:^30} {:^10} & ".format(formula_substring, '$\\star$'
                                                                  if doc.get('hull_distance') == 0 else ''))
            latex_struct_strings[-1] += ("{:^20.0f} & ".format(doc.get('hull_distance') * 1000)
                                         if doc.get('hull_distance', 0) > 0 else '{:^20} &'.format('-'))
            latex_struct_strings[-1] += "{:^20} & ".format(doc.get('space_group', 'xxx'))
            prov = get_guess_doc_provenance(doc['source'], doc.get('icsd'))
            if doc.get('icsd'):
                prov += ' {}'.format(doc['icsd'])
            latex_struct_strings[-1] += "{:^30} & ".format(prov)
            latex_struct_strings[-1] += "{:^30} \\\\".format('')

        if last_formula != formula_substring:
            if per_atom:
                gs_enthalpy = recursive_get(doc, energy_key)
            else:
                gs_enthalpy = recursive_get(doc, energy_key) * doc['num_atoms'] / doc['num_fu']

        last_formula = formula_substring

        if details:
            detail_string, detail_substring = _construct_detail_strings(doc, padding_length=len(header_string), source=source)
            detail_strings.append(detail_string)
            detail_substrings.append(detail_substring)

        if source:
            source_strings.append(_construct_source_string(doc['source']))

    total_string = ''
    total_string += len(header_string) * '─' + '\n'
    total_string += header_string + '\n'
    total_string += units_string + '\n'
    total_string += len(header_string) * '─' + '\n'

    if markdown:
        markdown_string += len(header_string) * '-' + '\n'
        markdown_string += header_string + '\n'
        markdown_string += units_string + '\n'
        markdown_string += len(header_string) * '-' + '\n'

    summary_inds = []
    # filter for lowest energy phase per stoichiometry
    if summary:
        current_formula = ''
        formula_list = {}
        for ind, substring in enumerate(formulae):
            if substring != current_formula and substring not in formula_list:
                current_formula = substring
                formula_list[substring] = 0
                summary_inds.append(ind)
            formula_list[substring] += 1
    else:
        summary_inds = range(len(struct_strings))

    # construct final string containing table
    if markdown:
        markdown_string += '\n'.join(struct_strings[ind] for ind in summary_inds)
    elif latex:
        latex_string += '\n'.join(latex_struct_strings[ind] for ind in summary_inds)
    else:
        for ind in summary_inds:
            total_string += struct_strings[ind] + '\n'
            if details:
                total_string += detail_strings[ind] + '\n'
                total_string += detail_substrings[ind] + '\n'
            if source:
                total_string += source_strings[ind] + '\n'
            if details or source:
                total_string += len(header_string) * '─' + '\n'

    if markdown:
        markdown_string += '```'
        return markdown_string

    if latex:
        latex_string += '\\end{tabular}'
        return latex_string

    if return_str:
        return total_string

    print(total_string)


def loading_bar(iterable, width=80, verbosity=0):
    """ Checks if tqdm exists and makes a loading bar, otherwise
    just returns initial iterable.

    Parameters:
        iterable (iterable): the thing to be iterated over.

    Keyword arguments:
        width (int): maximum number of columns to use on screen.

    Returns:
        iterable: the decorated iterator.

    """
    try:
        import tqdm
        if verbosity < 1:
            raise RuntimeError
        return tqdm.tqdm(iterable, ncols=width)
    except (ImportError, RuntimeError):
        return iterable


def set_cursor_from_array(cursor, array, key):
    """ Updates the key-value pair for documents in
    internal cursor from a numpy array.
    """
    if len(array) != len(cursor):
        raise RuntimeError('Trying to fit array of shape {} into cursor of length {}'
                           .format(np.shape(array), len(cursor)))
    for ind, _ in enumerate(cursor):
        recursive_set(cursor[ind], key, array[ind])


def get_array_from_cursor(cursor, key, pad_missing=False):
    """ Returns a numpy array of the values of a key
    in a cursor, where the key can be defined as list
    of keys to use with `recursive_get`.

    Parameters:
        cursor (list): list of matador dictionaries.
        key (str or list): the key to extract, or list
            of keys/subkeys/indices to extract with
            recursive_get.

    Keyword arguments:
        pad_missing (bool): whether to fill array with NaN's
            where data is missing.
    Raises:
        KeyError: if any document is missing that key,
            unless pad_missing is True.

    Returns:
        np.ndarray: numpy array containing results, padded
            with np.nan if key is missing and pad_missing is True.
    """
    array = []
    for ind, doc in enumerate(cursor):
        try:
            if isinstance(key, (tuple, list)):
                array.append(recursive_get(doc, key))
            else:
                array.append(doc[key])
        except KeyError as exc:
            print('{} missing  in entry {}, with source {}'.format(key, ind, doc.get('source')))
            if pad_missing:
                array.append(np.NaN)
            else:
                raise exc
    array = np.asarray(array)
    return array


def get_guess_doc_provenance(sources, icsd=None):
    """ Returns a guess at the provenance of a structure
    from its source list.

    Return possiblities are 'ICSD', 'SWAP', 'OQMD' or
    'AIRSS', 'MP' or 'PF'.
    """
    prov = 'AIRSS'
    if isinstance(sources, dict):
        sources = sources['source']
    elif isinstance(sources, str):
        sources = [sources]
    for fname in sources:
        fname_with_folder = fname
        fname = fname.split('/')[-1].lower()
        if (fname.endswith('.castep') or fname.endswith('.res') or fname.endswith('.history') or
                ('oqmd' in fname and fname.count('.') == 0) or
                (any(s in fname for s in ['mp-', 'mp_']) and fname.count('.') == 0) or
                (any(s in fname for s in ['pf-', 'pf_']) and fname.count('.') == 0)):
            if any(substr in fname for substr in ['collcode', 'colcode', 'collo']):
                if fname.count('-') == 2 + fname.count('oqmd') or 'swap' in fname:
                    prov = 'SWAPS'
                else:
                    prov = 'ICSD'
            elif 'swap' in fname_with_folder:
                prov = 'SWAPS'
            elif '-ga-' in fname:
                prov = 'GA'
            elif icsd is not None:
                prov = 'ICSD'
            elif 'oqmd' in fname:
                prov = 'OQMD'
            elif '-icsd' in fname:
                prov = 'ICSD'
            elif 'pf-' in fname and prov is None:
                prov = 'PF'
            elif any(s in fname for s in ['mp-', 'mp_']) and prov != 'PF':
                prov = 'MP'
            elif '-sm-' in fname:
                prov = 'SM'
            elif '-doi-' in fname:
                prov = 'DOI'
            elif '-config_enum' in fname:
                prov = 'ENUM'

    return prov


def filter_unique_structures(cursor, quiet=False, **kwargs):
    """ Wrapper for `matador.fingerprints.similarity.get_uniq_cursor` that
    displays the results and returns the filtered cursor.

    """
    from matador.fingerprints.similarity import get_uniq_cursor
    uniq_inds, dupe_dict, _, _ = get_uniq_cursor(cursor, **kwargs)
    filtered_cursor = [cursor[ind] for ind in uniq_inds]

    if not quiet:
        display_cursor = []
        additions = []
        deletions = []
        for key in dupe_dict:
            additions.append(len(display_cursor))
            display_cursor.append(cursor[key])
            if dupe_dict[key]:
                for _, jnd in enumerate(dupe_dict[key]):
                    deletions.append(len(display_cursor))
                    display_cursor.append(cursor[jnd])

        if not display_cursor:
            display_cursor = filtered_cursor

        display_results(
            display_cursor,
            additions=additions,
            deletions=deletions,
            sort=True,
            use_source=True,
            **kwargs
        )

    print('Filtered {} down to {}'.format(len(cursor), len(uniq_inds)))

    return filtered_cursor


def filter_cursor(cursor, key, vals, verbosity=0):
    """ Returns a cursor obeying the filter on the given key. Any
    documents that are missing the key will not be returned. Any
    documents with values that cannot be compared to floats will also
    not be returned.

    Parameters:
        cursor (list): list of dictionaries to filter.
        key (str): key to filter.
        vals (list): either 1 value to 2 values to use as a range.
            The values are interpreted as floats for comparison.

    Returns:
        list: list of dictionaries that pass the filter.

    """
    filtered_cursor = list()
    orig_cursor_len = len(cursor)
    if not isinstance(vals, list):
        vals = [vals]
    if len(vals) == 2:
        min_val = float(vals[0])
        max_val = float(vals[1])
        if verbosity > 0:
            print('Filtering {} <= {} < {}'.format(min_val, key, max_val))
        for doc in cursor:
            try:
                if doc[key] < max_val and doc[key] >= min_val:
                    filtered_cursor.append(doc)
            except (TypeError, ValueError, KeyError):
                pass
    else:
        min_val = float(vals[0])
        if verbosity > 0:
            print('Filtering {} >= {}'.format(key, min_val))
        for doc in cursor:
            try:
                if doc[key] >= min_val:
                    filtered_cursor.append(doc)
            except (TypeError, ValueError, KeyError):
                pass

    if verbosity > 0:
        print(orig_cursor_len, 'filtered to', len(filtered_cursor), 'documents.')
    return filtered_cursor


def filter_cursor_by_chempots(species, cursor):
    """ For the desired chemical potentials, remove any incompatible structures
    from cursor.

    Parameters:
        species (list): list of chemical potential formulae.
        cursor (list): list of matador documents to filter.

    Returns:
        list: the filtered cursor.

    """
    from matador.utils.chem_utils import get_number_of_chempots
    # filter out structures with any elements with missing chem pots
    chempot_stoichiometries = []
    for label in species:
        chempot_stoichiometries.append(get_stoich_from_formula(label))

    inds_to_remove = set()
    for ind, doc in enumerate(cursor):
        try:
            cursor[ind]['num_chempots'] = get_number_of_chempots(doc, chempot_stoichiometries)
        except RuntimeError:
            inds_to_remove.add(ind)
        else:
            cursor[ind]['concentration'] = (cursor[ind]['num_chempots'][:-1] /
                                            np.sum(cursor[ind]['num_chempots'])).tolist()
            for idx, conc in enumerate(cursor[ind]['concentration']):
                if conc < 0 + EPS:
                    cursor[ind]['concentration'][idx] = 0.0
                elif conc > 1 - EPS:
                    cursor[ind]['concentration'][idx] = 1.0

    return [doc for ind, doc in enumerate(cursor) if ind not in inds_to_remove]


def _construct_structure_string(
        doc, ind, formula_substring, gs_enthalpy, use_source, colour, hull, additions, deletions,
        add_index_mode, del_index_mode, energy_key, per_atom, eform, markdown, latex
):
    """ Construct the pretty output for an individual structure.

    Options passed from `matador.utils.cursor_utils.display_results.`

    Returns:
        str: the pretty output.

    """

    # start with two spaces, replaced by the prefix from hull/add/del
    this_struct_string = '  '
    prefix = ''
    suffix = ''
    # apply appropriate prefices and suffices to structure
    if hull and np.abs(doc.get('hull_distance')) <= 0.0 + 1e-12:
        if colour:
            prefix = '\033[92m'
            suffix = '\033[0m'
        this_struct_string = '* '

    if additions is not None:
        if (add_index_mode and ind in additions) or doc.get('text_id', '_') in additions:
            this_struct_string = '+ '
            if colour:
                prefix = '\033[92m'
                suffix = '\033[0m'
    if deletions is not None:
        if (del_index_mode and ind in deletions) or doc.get('text_id', '_') in deletions:
            this_struct_string = '- '
            if colour:
                prefix = '\033[91m'
                suffix = '\033[0m'

    # display the canonical name for the structure
    if use_source:
        src = get_root_source(doc['source'])
        max_len = 34
        this_struct_string += "{:<36.{max_len}}".format(
            src if len(src) <= max_len else src[:max_len-4]+'[..]', max_len=max_len
        )
    else:
        this_struct_string += "{:^24.22}".format(' '.join(doc.get('text_id', ['xxx', 'yyy'])))

    # again, if we're not outputting to markdown, then flag warnings in the quality column
    try:
        if doc.get('prototype'):
            this_struct_string += "{:^5}".format('*p*')
        elif doc.get('quality', 5) == 0:
            this_struct_string += "{:^5}".format('!!!')
        else:
            this_struct_string += "{:^5}".format((5 - doc.get('quality', 5)) * '?')
    except KeyError:
        this_struct_string += "{:^5}".format(' ')

    # loop over header names and print the appropriate values
    if 'pressure' in doc and doc['pressure'] != 'xxx':
        this_struct_string += "{: >9.2f} ".format(doc['pressure'])
    else:
        this_struct_string += "{:^9} ".format('xxx')

    try:
        if per_atom and 'cell_volume' in doc and 'num_atoms' in doc:
            this_struct_string += "{:>12.1f}  ".format(doc['cell_volume'] / doc['num_atoms'])
        elif 'cell_volume' in doc and 'num_fu' in doc:
            this_struct_string += "{:>12.1f}  ".format(doc['cell_volume'] / doc['num_fu'])
        else:
            this_struct_string += "{:^12}  ".format('xxx')
    except Exception:
        this_struct_string += "{:^10} ".format('xxx')

    try:
        if hull and eform:
            this_struct_string += "{:>12.3f}      ".format(
                doc['formation_' + energy_key]
            )
        elif hull:
            this_struct_string += "{:>12.1f}      ".format(
                1000 * doc['hull_distance']
            )
        elif per_atom:
            this_struct_string += "{:>16.4f}  ".format(recursive_get(doc, energy_key) - gs_enthalpy)
        else:
            this_struct_string += "{:>16.4f}  ".format(
                recursive_get(doc, energy_key) * doc['num_atoms'] / doc['num_fu'] - gs_enthalpy
            )
    except KeyError:
        this_struct_string += "{:^18}".format('xxx')

    if latex:
        from matador.utils.cell_utils import get_space_group_label_latex
        this_struct_string += " {:^13} ".format(get_space_group_label_latex(doc.get('space_group', 'xxx')))
    else:
        this_struct_string += " {:^13} ".format(doc.get('space_group', 'xxx'))

    # now we add the formula column
    this_struct_string += " {:^13} ".format(formula_substring)

    if 'num_fu' in doc:
        this_struct_string += " {:^6} ".format(int(doc['num_fu']))
    else:
        this_struct_string += " {:^6} ".format('xxx')

    if 'source' in doc:
        prov = get_guess_doc_provenance(doc['source'], doc.get('icsd'))
        this_struct_string += "{:^8}".format(prov)
    else:
        this_struct_string += "{:^8}".format('xxx')

    this_struct_string = prefix + this_struct_string + suffix

    return this_struct_string


def _construct_source_string(sources):
    """ From a list of sources, return a fancy string output
    displaying them as a list.

    """
    num_sources = len(sources)
    if num_sources == 1:
        this_source_string = 11 * ' ' + u"└──────────────────"
    else:
        this_source_string = 11 * ' ' + u"└───────────────┬──"

    for num, _file in enumerate(sources):
        if num_sources == 1:
            this_source_string += ''
        elif num == num_sources - 1:
            this_source_string += (len(u"└────────────── ") + 11) * ' ' + u'└──'
        elif num != 0:
            this_source_string += (len(u"└────────────── ") + 11) * ' ' + u'├──'

        this_source_string += ' ' + _file
        if num != num_sources - 1:
            this_source_string += '\n'

    return this_source_string


def _construct_detail_strings(doc, padding_length=0, source=False):
    """ From a document, return a fancy string output
    displaying the desired field names and details of
    the structure.

    Parameters:
        doc (dict): matador document to print.

    Keyword arguments:
        source (bool): whether to allow adjust output so it
            can be interleaved with the source pretty output.
        padding_length (int): how much to pad the detail string
            output.

    Returns:
        (str, str): containing pretty output over two lines.

    """
    detail_string = 11 * ' ' + u"├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌ "
    if source:
        detail_substring = 11 * ' ' + u"├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌ "
    else:
        detail_substring = 11 * ' ' + u"└╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌ "

    if 'spin_polarized' in doc:
        if doc['spin_polarized']:
            detail_string += 'S-'

    if 'sedc_scheme' in doc:
        detail_string += doc['sedc_scheme'].upper() + '+'

    if 'xc_functional' in doc:
        detail_string += doc['xc_functional']
    else:
        detail_string += 'xc-functional unknown!'

    if 'cut_off_energy' in doc:
        detail_string += ", {:4.2f} eV".format(doc['cut_off_energy'])
    else:
        detail_string += 'cutoff unknown'

    if 'external_pressure' in doc:
        detail_string += ", {:4.2f} GPa".format(
            sum(doc['external_pressure'][i][0] for i in range(3)) / 3
        )

    if 'kpoints_mp_spacing' in doc:
        detail_string += ", ~{:0.3f} 1/A".format(doc['kpoints_mp_spacing'])

    if 'geom_force_tol' in doc:
        detail_string += ', {:.2f} eV/A, '.format(doc['geom_force_tol'])

    if 'species_pot' in doc:
        for species in doc['species_pot']:
            detail_substring += '{}: {}, '.format(species, doc['species_pot'][species])

    if 'icsd' in doc:
        detail_substring += 'ICSD-CollCode {}, '.format(doc['icsd'])

    if 'tags' in doc:
        if isinstance(doc['tags'], list):
            detail_substring += ', '.join(doc['tags'])

    if 'user' in doc:
        detail_substring += doc['user']

    if 'encapsulated' in doc and all(key in doc for key in ['cnt_chiral', 'cnt_radius', 'cnt_length']):
        detail_string += (
            ', (n,m)=(' + str(doc['cnt_chiral'][0]) + ',' + str(doc['cnt_chiral'][1]) + ')'
        )
        detail_string += ", r={:4.2f} A".format(doc['cnt_radius'])
        detail_string += ", z={:4.2f} A".format(doc['cnt_length'])

    detail_string += ' ' + (padding_length - len(detail_string) - 1) * u"╌"
    detail_substring += ' ' + (padding_length - len(detail_substring) - 1) * u"╌"

    return detail_string, detail_substring


def _construct_header_string(markdown, use_source, per_atom, eform, hull, summary, energy_key):
    """ Construct the header of the table, from the passed options.
    For arguments, see docstring for `matador.utils.cursor_utils.display_results`.

    Returns:
        (str, str): the header and units string.

    """
    header_string = ''
    units_string = ''
    if not markdown:
        if use_source:
            header_string += "{:^38}".format('Source')
            units_string += "{:^38}".format('')
        else:
            header_string += "{:^26}".format('ID')
            units_string += "{:^26}".format('')

        header_string += "{:^5}".format('!?!')
        units_string += "{:^5}".format('')
    else:
        header_string += '```\n'
        header_string += "{:^43}".format('Root')
        units_string += "{:^43}".format('')

    header_string += "{:^10}".format('Pressure')
    units_string += "{:^10}".format('(GPa)')

    header_string += "{:^14}".format('Cell volume')
    if per_atom:
        units_string += "{:^14}".format('(Å³/atom)')
    else:
        units_string += "{:^14}".format('(Å³/fu)')

    if eform:
        header_string += "{:^18}".format('Formation energy')
        units_string += "{:^18}".format('(eV/atom)')
    elif hull:
        header_string += "{:^18}".format('Hull dist.')
        units_string += "{:^18}".format('(meV/atom)')
    elif per_atom:
        header_string += "{:^18}".format(' '.join(energy_key.replace('_per_atom', '').split('_')).title())
        units_string += "{:^18}".format('(eV/atom)')
    else:
        header_string += "{:^18}".format(' '.join(energy_key.replace('_per_atom', '').split('_')).title())
        units_string += "{:^18}".format('(eV/fu)')

    header_string += "{:^15}".format('Space group')
    header_string += " {:^13} ".format('Formula')
    header_string += "{:^8}".format('# fu')
    header_string += "{:^8}".format('Prov.')

    if summary:
        header_string += '{:^12}'.format('Occurrences')

    return header_string, units_string



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the Site class for handling
atomic sites.

"""

import numpy as np
import copy
from matador.utils.cell_utils import cart2frac, frac2cart, wrap_frac_coords


class Site:
    def __init__(self, species: str, position: list, lattice_cart,
                 position_unit='fractional', **site_data):

        self.lattice = copy.deepcopy(lattice_cart)
        self.set_position(position, position_unit)
        self.species = species
        self._occupancy = None
        self.site_data = {}

        if site_data.get('voronoi_substructure') is not None:
            assert self.species == site_data['voronoi_substructure'][0]
            self.site_data['voronoi_substructure'] = site_data['voronoi_substructure'][1]
            del site_data['voronoi_substructure']

        self.site_data.update(site_data)

    def __str__(self):
        site_str = '{species} {pos[0]:4.4f} {pos[1]:4.4f} {pos[2]:4.4f}'.format(species=self.species, pos=self.coords)
        for key in self.site_data:
            site_str += '\n{} = {}'.format(key, self.site_data[key])
        return site_str

    def __repr__(self):
        site_str = '{species} {pos[0]:4.4f} {pos[1]:4.4f} {pos[2]:4.4f}'.format(species=self.species, pos=self.coords)
        for key in self.site_data:
            site_str += '\n{} = {}'.format(key, self.site_data[key])
        return site_str

    def __deepcopy__(self, memo):
        from copy import deepcopy
        species, position, lattice = (deepcopy(x) for x in (self.species, self._coords['fractional'], self.lattice))
        site_data = deepcopy(self.site_data)
        return Site(species, position, lattice, position_unit='fractional', **site_data)

    def set_position(self, position, units):
        if len(position) != 3 or not all(isinstance(p, (float, int)) for p in position):
            raise RuntimeError('CrystalSite position has wrong shape: {}'.format(position))
        if '_coords' not in self.__dict__:
            self._coords = dict()
        if units == 'fractional':
            self._coords['fractional'] = wrap_frac_coords(
                [float(pos) for pos in position],
                remove=False
            )
            self._coords['cartesian'] = frac2cart(self.lattice, self.coords)
        elif units == 'cartesian':
            self._coords['cartesian'] = [float(pos) for pos in position]
            self._coords['fractional'] = wrap_frac_coords(
                cart2frac(self.lattice, self.coords(units='cartesian')),
                remove=False
            )
        else:
            raise RuntimeError('Unit system {} not understood, expecting `fractional`/`cartesian`'.format(units))

    @property
    def coords(self):
        return self._coords['fractional']

    def get_coords(self, units='fractional'):
        if units not in ['fractional', 'cartesian']:
            raise RuntimeError('Unit system {} not understood, expecting `fractional`/`cartesian`'.format(units))
        else:
            return self._coords[units]

    @property
    def occupancy(self):
        if self._occupancy is None:
            self._occupancy = self.site_data.get("site_occupancy", 1.0)
        return self._occupancy

    @property
    def coordination(self):
        if '_coordination' in self.__dict__:
            return self._coordination
        if self.voronoi_substructure is None:
            raise RuntimeError('Voronoi substructure not found.')

        coordination = {}
        eps = 0.05
        for atom, weight in self.voronoi_substructure:
            if weight >= 1 - eps:
                if atom not in coordination:
                    coordination[atom] = 1
                else:
                    coordination[atom] += 1
        self._coordination = coordination

        return coordination

    def displacement_between_sites(self, other_site):
        return np.asarray(self.get_coords(units='cartesian')) - np.asarray(other_site.get_coords(units='cartesian'))

    def distance_between_sites(self, other_site):
        return np.linalg.norm(
            self.displacement_between_sites(other_site)
        )



#!/usr/bin/env python

""" Some tests for high-throughput calculations. """

import unittest
import os
import shutil
import glob

import psutil
import numpy as np

from .utils import MatadorUnitTest, REAL_PATH, detect_program
from matador.compute import ComputeTask
from matador.scrapers import cell2dict, param2dict, phonon2dict


HOSTNAME = os.uname()[1]
PATHS_TO_DEL = ["completed", "bad_castep", "input", "logs", HOSTNAME]
VERBOSITY = 2
EXECUTABLE = "castep"


CASTEP_PRESENT = detect_program(EXECUTABLE)
MPI_PRESENT = detect_program("mpirun")

if CASTEP_PRESENT and MPI_PRESENT:
    NCORES = max(psutil.cpu_count(logical=False) - 2, 1)
else:
    NCORES = 1


@unittest.skipIf(not CASTEP_PRESENT, "CASTEP not found.")
class ElasticWorkflowTest(MatadorUnitTest):
    """ Run a spectral workflow calculation. """

    def test_bulk_mod(self):
        for _f in glob.glob(REAL_PATH + "data/elastic_workflow/*"):
            shutil.copy(_f, ".")

        cell_dict, _ = cell2dict("bulk_mod.cell", db=False)
        param_dict, _ = param2dict("bulk_mod.param", db=False)
        _ = ComputeTask(
            res="Si2.res",
            ncores=NCORES,
            nnodes=None,
            node=None,
            cell_dict=cell_dict,
            param_dict=param_dict,
            verbosity=VERBOSITY,
            compute_dir="/tmp/scratch_test",
            workflow_kwargs={"plot": False, "num_volumes": 3},
        )

        self.assertFalse(os.path.isfile("completed/Si2.bib"))
        self.assertFalse(os.path.isfile("completed/Si2.check"))

        self.assertTrue(os.path.isfile("completed/Si2.bulk_mod.results"))
        self.assertTrue(os.path.isfile("completed/Si2.bulk_mod.res"))
        self.assertTrue(os.path.isfile("completed/Si2.bulk_mod.castep"))
        self.assertFalse(os.path.isfile("completed/Si2.bulk_mod.png"))

        self.assertTrue(os.path.isfile("completed/Si2.res"))
        self.assertTrue(os.path.isfile("completed/Si2.geom"))
        self.assertTrue(os.path.isfile("completed/Si2.castep"))

        self.assertTrue(os.path.isfile("completed/Si2.bulk_mod.cell"))
        self.assertFalse(os.path.exists("/tmp/scratch_test"))
        self.assertFalse(os.path.exists("scratch_test_link"))


@unittest.skipIf(not CASTEP_PRESENT, "CASTEP not found.")
class PhononWorkflowTest(MatadorUnitTest):
    """ Run a spectral workflow calculation. """

    def test_phonon(self):
        for _f in glob.glob(REAL_PATH + "data/phonon_workflow/*"):
            shutil.copy(_f, ".")

        cell_dict, _ = cell2dict("Si.cell", db=False)
        param_dict, _ = param2dict("Si.param", db=False)
        _ = ComputeTask(
            res="Si2.res",
            ncores=NCORES,
            nnodes=None,
            node=None,
            cell_dict=cell_dict,
            param_dict=param_dict,
            verbosity=VERBOSITY,
            compute_dir="tmpier_tst",
        )

        self.assertFalse(os.path.isfile("completed/Si2.bib"))
        self.assertTrue(os.path.isfile("completed/Si2.check"))

        self.assertTrue(os.path.isfile("completed/Si2.bands"))
        self.assertTrue(os.path.isfile("completed/Si2.castep"))
        self.assertTrue(os.path.isfile("completed/Si2.phonon"))
        self.assertTrue(os.path.isfile("completed/Si2.phonon_dos"))

        phon, s = phonon2dict("completed/Si2.phonon")
        self.assertTrue(s, msg="Failed to read phonon file")
        self.assertGreater(np.min(phon["eigenvalues_q"]), -0.05)

        self.assertTrue(os.path.isfile("completed/Si2.cell"))
        self.assertTrue(os.path.isfile("completed/Si2.res"))


@unittest.skipIf(not CASTEP_PRESENT, "CASTEP not found.")
class SpectralWorkflowTest(MatadorUnitTest):
    """ Run a spectral workflow calculation. """

    def test_full_spectral_in_compute_dir(self):
        for _f in glob.glob(REAL_PATH + "data/spectral_workflow/*"):
            shutil.copy(_f, ".")

        cell_dict, _ = cell2dict("Si.cell", db=False)
        param_dict, _ = param2dict("Si.param", db=False)
        _ = ComputeTask(
            res="Si2.res",
            ncores=NCORES,
            nnodes=None,
            node=None,
            cell_dict=cell_dict,
            param_dict=param_dict,
            verbosity=VERBOSITY,
            compute_dir="tmpier_tst",
        )

        self.assertFalse(os.path.isfile("completed/Si2.bib"))
        self.assertTrue(os.path.isfile("completed/Si2.check"))

        self.assertTrue(os.path.isfile("completed/Si2-out.cell_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2-out.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2-out.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.adaptive.agr"))
        self.assertTrue(os.path.isfile("completed/Si2.adaptive.dat"))

        self.assertTrue(os.path.isfile("completed/Si2.bands"))
        self.assertTrue(os.path.isfile("completed/Si2.bands_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.bands_dos"))

        self.assertTrue(os.path.isfile("completed/Si2.castep"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cell"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cst_esp"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_pdos"))

        self.assertTrue(os.path.isfile("completed/Si2.odi"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.odo"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.param"))
        self.assertTrue(os.path.isfile("completed/Si2.param_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.param_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.param_scf"))
        self.assertTrue(os.path.isfile("completed/Si2.pdis.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.res"))

    def test_dos_only_spectral(self):
        for _f in glob.glob(REAL_PATH + "data/spectral_workflow/*"):
            shutil.copy(_f, ".")

        cell_dict, _ = cell2dict("Si.cell", db=False)
        del cell_dict["spectral_kpoints_path_spacing"]
        param_dict, _ = param2dict("Si.param", db=False)
        _ = ComputeTask(
            res="Si2.res",
            ncores=NCORES,
            nnodes=None,
            node=None,
            cell_dict=cell_dict,
            param_dict=param_dict,
            verbosity=VERBOSITY,
            compute_dir="tmpier_tst",
        )

        self.assertFalse(os.path.isfile("completed/Si2.bib"))
        self.assertTrue(os.path.isfile("completed/Si2.check"))

        self.assertFalse(os.path.isfile("completed/Si2-out.cell_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.bands_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.castep_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.cell_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.dome_bin_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.param_dispersion"))
        self.assertFalse(os.path.isfile("completed/Si2.pdos_bin_dispersion"))

        self.assertTrue(os.path.isfile("completed/Si2-out.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2-out.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.adaptive.agr"))
        self.assertTrue(os.path.isfile("completed/Si2.adaptive.dat"))

        self.assertTrue(os.path.isfile("completed/Si2.bands"))
        self.assertTrue(os.path.isfile("completed/Si2.bands_dos"))

        self.assertTrue(os.path.isfile("completed/Si2.castep"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cell"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cst_esp"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_pdos"))

        self.assertTrue(os.path.isfile("completed/Si2.odi"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_broadening"))
        self.assertFalse(os.path.isfile("completed/Si2.odi_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.odo"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_broadening"))
        self.assertFalse(os.path.isfile("completed/Si2.odo_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.param"))
        self.assertTrue(os.path.isfile("completed/Si2.param_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.param_scf"))
        self.assertFalse(os.path.isfile("completed/Si2.pdis.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_dos"))
        self.assertFalse(os.path.isfile("completed/Si2.pdos_bin_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.res"))

    def test_full_spectral(self):
        for _f in glob.glob(REAL_PATH + "data/spectral_workflow/*"):
            shutil.copy(_f, ".")

        cell_dict, _ = cell2dict("Si.cell")
        param_dict, _ = param2dict("Si.param", db=False)
        _ = ComputeTask(
            res="Si2.res",
            ncores=NCORES,
            nnodes=None,
            node=None,
            cell_dict=cell_dict,
            param_dict=param_dict,
            verbosity=VERBOSITY,
            compute_dir=None,
        )

        self.assertFalse(os.path.isfile("completed/Si2.bib"))
        self.assertTrue(os.path.isfile("completed/Si2.check"))

        self.assertTrue(os.path.isfile("completed/Si2-out.cell_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2-out.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2-out.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.adaptive.agr"))
        self.assertTrue(os.path.isfile("completed/Si2.adaptive.dat"))

        self.assertTrue(os.path.isfile("completed/Si2.bands"))
        self.assertTrue(os.path.isfile("completed/Si2.bands_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.bands_dos"))

        self.assertTrue(os.path.isfile("completed/Si2.castep"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.castep_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cell"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.cell_scf"))

        self.assertTrue(os.path.isfile("completed/Si2.cst_esp"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.dome_bin_pdos"))

        self.assertTrue(os.path.isfile("completed/Si2.odi"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odi_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.odo"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.odo_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.param"))
        self.assertTrue(os.path.isfile("completed/Si2.param_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.param_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.param_scf"))
        self.assertTrue(os.path.isfile("completed/Si2.pdis.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos.dat"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_broadening"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_dispersion"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_dos"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_pdis"))
        self.assertTrue(os.path.isfile("completed/Si2.pdos_bin_pdos"))
        self.assertTrue(os.path.isfile("completed/Si2.res"))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest
import numpy as np
import json


class MiscUtilTest(unittest.TestCase):
    """ Tests cursor util functions. """

    def test_serialize_numpy(self):
        from matador.utils.print_utils import dumps

        doc = {"a": "b", "1": [1, 2, 3], "2": np.zeros((3, 3))}
        doc_list = {"a": "b", "1": [1, 2, 3], "2": [[0, 0, 0], [0, 0, 0], [0, 0, 0]]}
        string = dumps(doc)
        self.assertEqual(json.loads(string), doc_list)


if __name__ == "__main__":
    unittest.main()
